import logging
from collections import defaultdict
from typing import List
import numpy as np
import pandas as pd
from base import Base
from plots.plots_interrogate_mode import *
from predictions.apply import apply_tool_predictions
from predictions.metrics import generate_statistics
from preprocessing.osutils import check_file_exists
from preprocessing.osutils import ensure_folder_exists


class PredictionsEval(Base):
    """
    Inspect predictions of an unlabelled VCF
    """

    def __init__(self, vcf: str,
                 out_dir: str,
                 scopes_to_predict: List = None,
                 types_of_variant: List = None,
                 metric: str = "weighted_accuracy",
                 location: str = "HGVSc",
                 aggregate_classes: bool = False,
                 select_conseqs: str = "in_gene_body",
                 do_intronic_analysis: bool = False,
                 plot_these_tools: List = None,
                 labels: str = None,
                 allele_frequency_col: str = "gnomADg_AF",
                 skip_heatmap: bool = False,
                 tools_config: str = None,
                 interrogate_mode: bool = False):

        """
        ----
        Base args described in Base class
        ----
        :param List plot_these_tools: Plot scores distribution for the
            given tools. Default: `None`.

        :param str labels: If variants stored in VCF represent a list
            of labelled variants of a given type. If so, additional
            metrics will be inferred. Available types: ['benign',
            'pathogenic']. Default: `None`, variants are unlabelled.
        """

        super().__init__(vcf=vcf,
                         out_dir=out_dir,
                         scopes_to_predict=scopes_to_predict,
                         types_of_variant=types_of_variant,
                         metric=metric,
                         location=location,
                         aggregate_classes = aggregate_classes,
                         select_conseqs = select_conseqs,
                         do_intronic_analysis=do_intronic_analysis,
                         is_clinvar=False,
                         allele_frequency_col=allele_frequency_col,
                         skip_heatmap=skip_heatmap,
                         tools_config=tools_config,
                         interrogate_mode=interrogate_mode)

        if plot_these_tools is not None:
            assert set(plot_these_tools).issubset(self.available_tools), "Invalid tool(s) name(s) in the " \
                                                                         "'--plot_these_tools' argument. " \
                                                                         "List of available tools:\n{}". \
                format(self.available_tools)

        if "F1" in self.metric and self.labels is not None:
            raise ValueError("F1-based metrics are not available in the "
                             "interrogate mode, since all variants refer to a "
                             "single class type (--label set to {})".format(self.labels))

        self.plot_these_tools = plot_these_tools
        self.labels = labels

        self.inspect_predictions()

    def inspect_predictions(self):
        """
        Inspect predictions for each
        variant type and produces output
        files and plots

        :return:
        """
        def update_hgvsc(row: pd.Series):
            try:
                return row.SYMBOL + ":" + row.HGVSc.split(":")[1]
            except IndexError:
                if not row.SYMBOL or row.SYMBOL == "":
                    return row.HGVSg
                else:
                    return row.SYMBOL + ":" + row.HGVSg
                    

        df_with_pred = apply_tool_predictions(self.df, self.thresholds). \
            rename(columns={'index': 'variant_id'})

        for var_type, _func in self.variant_types:

            outdir = os.path.join(self.out_dir, var_type)
            ensure_folder_exists(outdir)
      
            _df_by_type = _func(df_with_pred).copy()

            logging.info("Looking at {} ({} variants)".format(var_type,
                                                              _df_by_type.shape[0]))

            if _df_by_type.shape[0] == 0:
                logging.warning("WARN: There are no {} in the variant set. "
                                "Skipping this analysis.".format(var_type))
                continue
            

            ###################
            ### Predictions ###
            ###################
            # df = apply_tool_predictions(df, threshold_list)
            logging.info("Extracting predictions")

            to_remove_cols = ['variant_id', 'type', 'subtype', 'Existing_variation', 'location', 'label', 'outcome']
            out_c = ['chr', 'pos', 'ref', 'alt', 'Consequence', 'HGVSc', 'HGVSg', 'SYMBOL'] + [col for col in _df_by_type.columns if '_prediction' in col]
            out_c_raw = [col for col in _df_by_type.columns if not '_prediction' in col and col not in to_remove_cols]

            out_df = _df_by_type[out_c]
            out_df_raw = _df_by_type[out_c_raw]
            
            # Add used threshold to the header
            tool_idx = {t[0]: i for i, t in enumerate(self.thresholds)}
            
            for i, _cols in enumerate([out_c, out_c_raw]):
                tool_rename = {}
                for c in _cols:
                    tool = c.split('_')[0]
                    if tool in tool_idx.keys():
                        direction = str(self.thresholds[tool_idx[tool]][1])
                        thresh = str(self.thresholds[tool_idx[tool]][2])
                        tool_rename[c] = tool + ' ({}{})'.format(direction, thresh)
                    else:
                        tool_rename[c] = c
                    
                if i == 0:
                    out_df = out_df.rename(columns=tool_rename) 
                    out_df.to_csv(os.path.join(outdir, "individual_predictions.tsv"), 
                                  index=False, 
                                  sep="\t")
                else:
                    out_df_raw = out_df_raw.rename(columns=tool_rename)
                    out_df_raw.to_csv(os.path.join(outdir, "individual_predictions_raw_scores.tsv"),index=False,
                                      sep="\t")
            
            _df_just_pred = pd.concat([_df_by_type['variant_id'],
                                       _df_by_type[[col for col in _df_by_type.columns if '_prediction' in col]],
                                       _df_by_type["HGVSc"],
                                       _df_by_type["HGVSg"],
                                       _df_by_type["SYMBOL"],
                                       _df_by_type["location"].to_frame()],
                                      axis=1)
                    
            _df_just_pred['HGVSc'] = _df_just_pred.apply(update_hgvsc, axis=1)
            _df_just_pred = _df_just_pred.rename(columns={col: col.split('_prediction')[0]
                                                          for col in _df_just_pred.columns})

            _df_just_pred['tools_preds'] = _df_just_pred.apply(lambda row: row[row == True].index.tolist(), axis=1)
            _df_just_pred['tools_preds'] = _df_just_pred['tools_preds'].apply(lambda x: ';'.join(x))
            any_patho = _df_just_pred[_df_just_pred.tools_preds.apply(lambda x: len(x)) > 0]
            if any_patho.shape[0] > 0:
                any_patho[['variant_id', 'HGVSc', 'HGVSg', 'tools_preds']].to_csv(os.path.join(outdir, 'variants_with_any_pathogenic_pred.tsv'), index=False, sep="\t")
            else:
                logging.info('No variant had any pathogenic prediction by any tool')
            _df_just_pred.drop(columns='tools_preds', inplace=True)
            ##############
            ### Ratios ###
            ##############
            #try:
            logging.info("Inspecting variants for which a large "
                        "fraction of tools predicts pathogenicity.")


            ratios_df = _df_just_pred.set_index('HGVSc').drop(["location", "variant_id", "HGVSg", "SYMBOL"], axis=1).copy()

            n_tools = len(list(ratios_df))
            patho_counts = ratios_df.eq(True).sum(axis=1)
            benign_counts = ratios_df.eq(False).sum(axis=1)
            unpredictable_counts = ratios_df.fillna(-1).eq(-1).sum(axis=1)
            ratios_df = pd.concat([patho_counts, benign_counts, unpredictable_counts], axis=1)
            ratios_df.columns = ['Is pathogenic', 'Is benign', 'Unpredictable']
     
            ratios_df = ratios_df / n_tools
            #ratios_df = self._fix_col_names(ratios_df)

            _top_predicted_patho = ratios_df[ratios_df['Is pathogenic'] > 0.5]
      
            _top_predicted_benign = ratios_df[ratios_df['Is benign'] > 0.5]
            if _top_predicted_patho.shape[0] > 0:
                _top_predicted_patho.to_csv(os.path.join(outdir, "top_variant_candidates.tsv"), sep="\t")

            plot_area(ratios_df, outdir)
        
            ratios_df['Unpredictable'] *= 100
    
            plot_heatmap(ratios_df, outdir)
            plot_heatmap(_top_predicted_patho, 
                            outdir, 
                            display_annot=True)

            plot_heatmap(_top_predicted_patho, 
                        outdir, 
                        display_annot=True,
                        benign_too = _top_predicted_benign)
                
            # except KeyError:
            #     logging.info("No tool has predictions for the given variant type ({}), "
            #                  "analysis is going to be skipped.".format(var_type))
            # if VCF refers to variants of a
            # given label, compute additional
            # metrics
            ###################
            ### Performance ###
            ###################
            if self.labels is not None:
                logging.info("Inspecting tools performance based on the label provided ({})".format(self.labels))
                _df_just_pred['label'] = False if self.labels in ["Benign", "Neutral"] else True
                self.generate_performance_with_label(_df_just_pred,
                                                     outdir=outdir,
                                                     var_type=var_type)


            ###############
            ### Heatmap ###
            ###############
            if self.skip_heatmap is False:
                logging.info("Generating heatmap")
                class_dict = {True: 1, False: 0, np.nan: -1}
                _df = _df_just_pred.replace(class_dict)
                
                plot_heatmap_toptools(_df, filters=self.location_filters, outdir=outdir)

            ##################
            ### Score dist ###
            ##################
            if self.plot_these_tools is not None:
                logging.info("Generating score distribution plots")
                for tool in self.plot_these_tools:
                    # uses original df that contains
                    # scores, not just binary predictions
                    plot_tool_score_distribution(_df_by_type,
                                                 tool=tool,
                                                 thresholds=self.thresholds,
                                                 outdir=outdir)

        if self.do_intronic_analysis:
            logging.info("For now, intronic analysis is not available "
                         "in the inspect mode. Skipping it.")

    def generate_performance_with_label(self,
                                        df_pred: pd.DataFrame,
                                        outdir: str,
                                        var_type:str):

        """
        Evaluate tools performance considering that the
        intput VCF refers to a list of variants with a
        single and known label (e.g. benign or pathogenic)

        :param pd.DataFrame df_pred: Df with predictions
            for each variant
        :param str outdir: Output directory
        :param var_type: Variant types analysed 

        :return:
        """
        assert "F1" not in self.metric, "Can't use F1-based metrics in the inspect mode since it is " \
                                        "necessary to have at least two classes for its calculation." \
                                        "at least "

        os.makedirs(os.path.join(outdir, 'out_performance'), exist_ok=True)
        outdir = os.path.join(outdir, 'out_performance')
        for _location in self.location_filters:

            outfile = os.path.join(outdir, "statistics_{}_{}.tsv").format(var_type, _location)
            statistics = defaultdict(list)

            if _location == "all":
                df = df_pred.copy()
            else:
                df = df_pred[df_pred.location == _location].copy()
                
            if df.empty:
                logging.info("WARN: Input VCF does not have any '{}' variants. "
                             "Skipping label performance analysis.".format(_location))
                continue

            for tool, *args in self.thresholds:
                try:
                    if np.sum(~df[tool].isnull()) == 0:
                        continue
                    generate_statistics(df, statistics, _location, tool, is_single_label=True)

                except KeyError:
                    continue

                stats_df = pd.DataFrame(statistics).drop(columns=['filter'])
                stats_df.sort_values([self.metric], ascending=False).to_csv(outfile, sep="\t", index=False)
                plot_accuracy(stats_df, self.metric, _location, outdir)

    def _fix_col_names(self, ratios_df: pd.DataFrame):
        """
        Fixes colnames when unpredictable variants were not found

        :param pd.DataFrame ratios_df: Input df
        """

        if "Unpredictable" not in ratios_df.columns:
            for i in range(0, len(ratios_df.columns)):
                if ratios_df.columns[i] not in ['Is benign', 'Is pathogenic']:
                    ratios_df = ratios_df.rename(columns={ratios_df.columns[i]: "Unpredictable"})

        if "Unpredictable" not in ratios_df.columns:
            ratios_df['Unpredictable'] = 0.0
            
        return ratios_df




import os
from collections import defaultdict
from posixpath import split
from typing import Union, List, Tuple

import hgvs
import pandas as pd
from importlib import resources
import config

from predictions.filters import update_thresholds, subset_toolset_by_scope, \
    subset_variants_by_type, _extract_possible_filters
from preprocessing.clinvar import get_clinvar_cached, remove_clinvar_useless
from preprocessing.location import *
from preprocessing.osutils import check_file_exists, setup_output_directory, ensure_folder_exists
from preprocessing.utils_tools import *
from preprocessing.vcf import process_vcf

TOOLS_CONFIG = resources.open_text(config, 'tools_config.txt')

class Base(object):
    """
    Base class to parse input args and
    create the dataframe for downstream
    analysis
    """

    def __init__(self, vcf: Union[Tuple, str],
                 out_dir: str,
                 scopes_to_predict: List = None,
                 types_of_variant: List = None,
                 metric: str = "weighted_accuracy",
                 location: str = "HGVSc",
                 aggregate_classes: bool = False,
                 select_conseqs: str = "gene_body",
                 do_intronic_analysis: bool = False,
                 split_splice_sites: bool = False,
                 is_clinvar: bool = False,
                 allele_frequency_col: str = "gnomADg_AF",
                 skip_heatmap: bool = False,
                 tools_config: str = TOOLS_CONFIG,
                 interrogate_mode: bool = False):

        """
        :param Union[Tuple, str] vcf: Input VCF(s) to analyse.
        :param str out_dir: Path to the output directory.

        :param List scopes_to_predict: Restrict analysis
            to a subset of tools based on their scope.
            Available options: ['Conservation', 'Whole_genome',
            'Protein', 'Splicing']. Default: `None` tools
            from all scopes are used.

        :param List types_of_variant: Restrict analysis to the
            given variant types. Available options: ['snps',
            'indels', 'insertions', 'deletions', 'mnps'].
            Default: `None`, Performance is measured for all
            types together and for each subtype separately.

        :param str metric: Metric to rank the tools.
            Available options: ['weighted_accuracy', 'accuracy',
            'F1', 'weighted_F1', 'coverage']. Default:
            `weighted_accuracy`.

        :param str location: VEP annotated VCF field to
            extract location of the variants. Available
            options: ['HGVSc, 'Consequence']. Default: "HGVSc".

        :param bool aggregate_classes: Aggregate coding and splice region
            variant classes into more high-level concepts to be analyzed 
            together
        
        :paran str select_conseqs: How to select the top consequence
            block per variant
        
        :param bool do_intronic_analysis: Whether additional analysis of
            intronic variants extracted from HGVSc field will
            be performed
            
        :param bool split_splice_sites: Do separate analysis for donor and 
        acceptor variants when 'do_intronic_analysis' is set
            
        :param bool is_clinvar: Whether `vcf` is from Clinvar.
            Default: `False`

        :param str allele_frequency_col: VCF field that measures
            population frequency of variants. If not present,
            analysis that depend of the field will be ignored.
            Default: `gnomADg_AF`

        :param bool skip_heatmap: Whether drawing of heatmaps
            should be skipped. Default: `False`

        :param str tools_config: Path to the tools config
            where available tools in VCF are mapped to the
            corresponding VCF annotation
            
        :param bool interrogate_mode: Whether veta run
            is in interrogate mode
            
        :return pd.DataFrame: Processed dataframe
        """
        self.metric = metric
        self.location_from = location
        self.do_intronic_analysis = do_intronic_analysis
        self.split_splice_sites = split_splice_sites
        self.aggregate_classes = aggregate_classes
        self.select_conseqs = select_conseqs
        self.location_filters = _extract_possible_filters(self.aggregate_classes)

        self.is_clinvar = is_clinvar
        self.is_interrogate_mode = interrogate_mode
        self.out_dir = setup_output_directory(out_dir)
        ensure_folder_exists(self.out_dir)
        self.allele_frequency_col = allele_frequency_col
        self.skip_heatmap = skip_heatmap

        tools_config = self._parse_tools_config(tools_config)
        self.thresholds = update_thresholds(tools_config)
        self.thresholds = subset_toolset_by_scope(self.thresholds, scopes_to_predict)
        self.available_tools = [t[0] for t in self.thresholds]

        # TODO change global thresholds list to remove excess of fields when custom models are provided
        self.tools_config = {k: v for k, v in tools_config.items() if k in self.available_tools}
        self.variant_types = subset_variants_by_type(types_of_variant)

        if self.location_from == "Consequence" and self.do_intronic_analysis:
            raise ValueError('If \'--do_intronic_analysis\' is set, \'--location\' '
                             'must be \'HGVSc\' because intronic bin are assigned '
                             'based on the distance offset extracted from HGVSc expressions.')

        if isinstance(vcf, str):
            check_file_exists(vcf)
            # if it's clinvar, additional
            # processing will be done
            self.df = self.get_df_ready(vcf)

        else:
            _dfs = []
            for f in vcf:
                is_benign = False if "pathogenic" in f or "deleterious" in f else True
                _dfs.append(self.get_df_ready(f, is_benign=is_benign))

            self.df = pd.concat(_dfs).reset_index(drop=True)

        # Replace missing AF with 0 for AF plots
        if self.allele_frequency_col in self.df.columns:

            if self.df[self.allele_frequency_col].dtype == object:
                self.df[self.allele_frequency_col] = self.df[self.allele_frequency_col].\
                    apply(lambda x: [0 if v is None else x[0] for v in x][0])

        # Replace missing AF if AF is used as a tool itself to evaluate performance
        if self.allele_frequency_col in [v[0][0] for v in self.tools_config.values()]:
            name = [tool for tool, v in self.tools_config.items() if v[0][0] == self.allele_frequency_col][0]
            # If tool name is different than vcf field (it it is the same, it's dealt above)
            try:
                if name != self.allele_frequency_col:
                    self.df[name] = self.df[name].fillna(0)
            except KeyError:
                raise KeyError('Unknown column {} in the processed dataframe. '
                               'Remove *tsv file and run again Clinvar analysis '
                               'with the added tool.'.format(name))

        if self.is_clinvar:
            # cache processed df
            self.df.to_csv(vcf + '.tsv', index=False)

    def get_df_ready(self, vcf: str, is_benign: bool = False):
        """
        Process input VCF(s) so that df are ready for
        downstream analysis

        :param str vcf: Input VCF file
        :param bool is_benign: Whether `vcf` refers to a set of
            benign variants. Only useful when `vcf` is not from
            Clinvar. Default: `False`
        :return pd.DataFrame: Cleaned dataframe
        """
        logging.info("Processing {} file".format(os.path.basename(vcf)))

        if self.is_clinvar:
            _df = get_clinvar_cached(vcf)
            # If cached file was found, DF is ready
            if _df is not None:
                return _df

        df, fields_missing_in_vcf = process_vcf(vcf, 
                                                tools_config=self.tools_config,
                                                thresholds=self.thresholds,
                                                select_conseq = self.select_conseqs,
                                                is_clinvar=self.is_clinvar,
                                                af_col=self.allele_frequency_col)

        # Remove tool from config and thresholds if corresponding field is absent from at least one of the VCFs
        self.tools_config = {k: v for k, v in self.tools_config.items() if k not in fields_missing_in_vcf}
        self.thresholds = [x for x in self.thresholds if x[0] not in fields_missing_in_vcf]

        if self.is_clinvar:
            df = df.dropna(subset=["CLNSIG", "CLNREVSTAT"])
            df = remove_clinvar_useless(df)
            logging.info("Number of variants after removing "
                         "irrelevant CLNSIG values: {}".format(df.shape[0]))

        # Extract variants
        logging.info("Assigning variants location/type")
        if self.location_from == "HGVSc":
            hp = hgvs.parser.Parser()

            df['location'] = df.apply(get_location,
                                      hp=hp,
                                      aggregate=self.aggregate_classes,
                                      axis=1,
                                      result_type='expand')
            
            if self.do_intronic_analysis:
                df[['intron_bin', 'intron_offset', 'which_ss']] = df.apply(
                    lambda x: assign_intronic_bins(x['HGVSc'], 
                                                   hp, 
                                                   x['location'], 
                                                   self.aggregate_classes), axis=1)

        elif self.location_from == "Consequence":
            df['location'] = df['Consequence'].apply(get_location_from_consequence, 
                                                     detailed=self.detailed_conseqs)

        # Clean predictions scores
        df = self._clean_scores(df.reset_index())

        # Set labels
        if self.is_clinvar:
            df['label'] = (df['CLNSIG'].str.contains('Pathogenic')) | \
                          (df['CLNSIG'].str.contains('Likely_pathogenic'))

        else:
            df['label'] = False if is_benign else True

        booleanDictionary = {True: 'Pathogenic', False: 'Benign'}
        df["outcome"] = df["label"].map(booleanDictionary)
    
        return df

    def _clean_scores(self, df: pd.DataFrame):
        """
        Transform tools scores so that proper
        evaluations can be performed.

        By default, VETA has some methods
        ready to process the scores of specific
        tools.

        :param pd.DataFrame df: Input dataframe
        with information about each variant.

        :return pd.DataFrame: Processed df.
        """

        # pd.options.display.float_format = '{:,.6f}'.format

        # values: [function_name, get_max, get_absolute_value]
        available_functions = {'to_numeric': [to_numeric, None, None],
                               'top_max': [get_top_pred, True, True],
                               'top_min': [get_top_pred, False, False],
                               'top_min_abs': [get_top_pred, False, True],
                               'categorical_to_numeric': [categorical_to_numerical, None, None],
                               'carol_like': [process_condel_carol, None, None],
                               'kipoi_like': [process_kipoi_tools, None, None],
                               'spliceai': [process_spliceai, None, None],
                               'trap': [process_trap, None, None],
                               'scap': [process_scap, None, None],
                               'conspliceml': [process_conspliceml, None, None],
                               'conspliceml_like': [process_conspliceml_like, None, None],
                               'pangolin': [process_pangolin, None, None],
                               'spip': [process_spip, None, None],
                               'labranchor': [process_labranchor, None, None],
                               'absplice_dna': [process_absplice, None, None]
                               }

        logging.info("Engineering the scores.")
        clean_functions = {
            "GERP": available_functions['top_max'],
            "phyloP": available_functions['top_max'],
            "phastCons": available_functions['top_max'],
            "SiPhy": available_functions['to_numeric'],
            "CDTS": available_functions['to_numeric'],

            "LRT": available_functions['to_numeric'],
            "Sift": available_functions['top_min_abs'],
            "Polyphen2HVAR": available_functions['top_max'],
            "Polyphen2HDIV": available_functions['top_max'],
            "MutationAssessor": available_functions['top_max'],
            "MutationTaster2": available_functions['top_max'],
            "FATHMM": available_functions['top_min'],
            "Provean": available_functions['top_min'],
            "Mutpred": available_functions['to_numeric'],
            "CAROL": available_functions['carol_like'],
            "Condel": available_functions['carol_like'],
            "M-CAP": available_functions['to_numeric'],
            "MetaLR": available_functions['to_numeric'],
            "MetaSVM": available_functions['to_numeric'],
            "REVEL": available_functions['top_max'],
            "VEST4": available_functions['to_numeric'],
            "CardioBoost": available_functions['to_numeric'],
            "CardioVAI": available_functions['to_numeric'],
            "PrimateAI": available_functions['to_numeric'],
            "VARITY": available_functions['to_numeric'],
            "MutFormer": available_functions['to_numeric'],
            "MutScore": available_functions['to_numeric'],
            "MVP": available_functions['to_numeric'],
            "MTR": available_functions['to_numeric'],
            "MPC": available_functions['to_numeric'],
            "MISTIC": available_functions['top_max'],
            "ClinPred": available_functions['top_max'],
            "cVEP": available_functions['categorical_to_numeric'],
            "EVE_Class20": available_functions['categorical_to_numeric'],
            "EVE_Class50": available_functions['categorical_to_numeric'],
            "EVE_Class90": available_functions['categorical_to_numeric'],
            "EVE": available_functions['to_numeric'],
            
            "fitCons": available_functions['to_numeric'],
            "LINSIGHT": available_functions['to_numeric'],
            "GWAVA": available_functions['to_numeric'],
            "CADD_v1.5": available_functions['to_numeric'],
            "CADD-Splice": available_functions['to_numeric'],
            "Eigen": available_functions['to_numeric'],
            "FATHMM-MKL": available_functions['to_numeric'],
            "FunSeq2": available_functions['top_max'],
            "DANN": available_functions['to_numeric'],
            "ReMM": available_functions['to_numeric'],
            "CAPICE": available_functions['top_max'],

            "HAL": available_functions['kipoi_like'],
            "S-CAP": available_functions['scap'],
            "MMSplice": available_functions['kipoi_like'],
            "kipoiSplice4": available_functions['kipoi_like'],
            "kipoiSplice4_cons": available_functions['kipoi_like'],
            "TraP": available_functions['trap'],
            "SPIDEX": available_functions['to_numeric'],
            "dbscSNV": available_functions['top_max'],
            "MaxEntScan": available_functions['top_max'],
            "SpliceAI": available_functions['spliceai'],
            "SQUIRLS": available_functions['to_numeric'],
            "ConSpliceML": available_functions['conspliceml'],
            "IntSplice2": available_functions['to_numeric'],
            "CI-SpliceAI": available_functions['spliceai'],
            "Pangolin": available_functions['pangolin'],
            "SPiP": available_functions['spip'],
            "MLCsplice": available_functions['top_max'],
            "AbSplice-DNA": available_functions['absplice_dna'],
            "LaBranchoR": available_functions['labranchor']
        }

        _functions_that_require_loc = ["process_trap"]
        _functions_that_require_symbol = ["process_spliceai", "process_conspliceml"]
        _absent_tools = {_tool: info for _tool, info in self.tools_config.items()
                         if _tool not in clean_functions.keys()}

        # If new custom tool
        if _absent_tools:
 
            # get function, if provided (5th col in config)
            for _tool, info in _absent_tools.items():
                if len(info) > 4:    
                    _function = info[4]
                    #self.tools_config[_tool] = info[:-1]
                else:
                    logging.info('Processing function was not provided for {}. Using the default to_numeric function. Be careful.'.format(_tool))
                    _function = 'to_numeric'
                clean_functions[_tool] = available_functions[_function]
        # For each tool belonging to the given scope
        for _tool, info in self.tools_config.items():
            # if it is known tool
            if _tool in clean_functions.keys():
                # get function to apply
                _f = clean_functions[_tool][0]
                # is_max
                is_max = clean_functions[_tool][1]
                # is_absolute
                is_absolute = clean_functions[_tool][2]

                # if the following tools, use custom methods
                # that need variant location
                if _f.__name__ in _functions_that_require_loc:
                    df[_tool] = df[[_tool] + ['location']].apply(_f, axis=1)

                elif _f.__name__ in _functions_that_require_symbol:
                    
                    # If CI-SpliceAI
                    if _f.__name__ == "process_spliceai" and _tool != "SpliceAI":
                        df[_tool] = df[[_tool] + ['SYMBOL']].apply(_f, check_gene_name=False, axis=1)
                    else:
                        df[_tool] = df[[_tool] + ['SYMBOL']].apply(_f, axis=1)
                    
                # apply single function to
                # the target columns
                else:
                    # if max and absolute info are not
                    # required (e.g. to_numeric, and custom
                    # function to specific tools)
                    if all([_v is None for _v in [is_max, is_absolute]]):
                        df[_tool] = df[_tool].apply(_f)
                                   
                    else:
                        df[_tool] = df[_tool].apply(_f, is_max=is_max,
                                                    absolute=is_absolute)

        return df

    def _parse_tools_config(self, config):
        """
        Parse tools config file to ensure
        correctness for downstream analysis.
        - Empty and commented lines (^#) are ignored.
        - First two fields are required
        - If tool name (1st col) not in the list
        of available tools, it expects to be a new
        custom method so that, 3rd (scope) and 4th
        and 5th columns are required.

        :param str config: Input file
        :return defaultdict: Dict with
        info about each tool
        """
        AVAILABLE_TOOLS_NAMES = ["GERP", "phyloP", "phastCons", "SiPhy",
                                 "CDTS", "LRT", "Sift", "Polyphen2HVAR",
                                 "Polyphen2HDIV", "MutationTaster2",
                                 "MutationAssessor", "FATHMM", "Provean",
                                 "Mutpred", "CAROL", "Condel", "M-CAP",
                                 "MetaSVM", "MetaLR", "REVEL", "VEST4",
                                 "MVP", "PrimateAI", "cVEP", "VARITY",
                                 "MTR", "MPC", "MutScore", "MutFormer",
                                 "ClinPred", "MISTIC",
                                 "EVE", "EVE_class20", "EVE_class50","EVE_class90",                             
                                 "CardioBoost", "CardioVAI",
                                 "fitCons", "LINSIGHT", "ReMM", "GWAVA", "FATHMM-MKL",
                                 "Eigen", "FunSeq2", "CADD_v1.5", "CADD-Splice", "DANN", "CAPICE",
                                 "HAL", "SPIDEX", "dbscSNV", "MaxEntScan", "SpliceAI", "S-CAP", "ConSpliceML",
                                 "TraP", "MMSplice", "SQUIRLS", "IntSplice2", "kipoiSplice4", "kipoiSplice4_cons",
                                 "CI-SpliceAI", "Pangolin", "SPiP", "LaBranchoR", "MLCsplice", "AbSplice-DNA"]

        AVAILABLE_SCOPES = ["Protein", "Conservation", "Whole_genome", "Splicing"]
        AVAILABLE_FUNCTIONS = ['to_numeric', 'top_max', 'top_min', 'top_min_abs',
                               'categorical_to_numeric', 'kipoi_like', 'carol_like', 
                               'trap', 'scap', 'conspliceml', 'conspliceml_like',
                               'spliceai', 'ci_spliceai',
                               'pangolin', 'spip', 'labranchor']

        data = defaultdict(list)
        if isinstance(config, str):
            _c = open(config, 'r')
            config = _c

        for line in config:
            line = line.rstrip()
            if line.startswith("#") or not line:
                continue

            _fields = line.split("\t")

            assert len(_fields) >= 2, "First 2 fields are required in the tools config file. " \
                                      "Problematic line: {}.\nMake sure fields are tab separated.".format(line)

            assert _fields[0] not in data.keys(), "Repeated tool name found in the tools " \
                                                  "config file: {}".format(_fields[0])

            data[_fields[0]].append(_fields[1].split(","))
            if _fields[0] not in AVAILABLE_TOOLS_NAMES:
                assert len(_fields) >= 5, "\'{}\' tool is not in the list of available " \
                                          "tools. Thus, It is assumed to be a custom method " \
                                          "provided by the user, which means that a reference " \
                                          "threshold, its directionality and scope must be provided. " \
                                          "Please set that info (at least 5 columns), or check if the " \
                                          "tool name was just misspelled. List of default tools: " \
                                          "{}".format(_fields[0], '\n' + '\n'.join(AVAILABLE_TOOLS_NAMES))

            if len(_fields) > 2:
                assert _fields[2] in ['>', "<"], "Directionality of the threshold (3th col) in " \
                                                 "the tools config must be set to '>' or '<' at " \
                                                 "line:\n{}".format(line)

                try:
                    _ref_thresh = float(_fields[3])
                except ValueError:
                    raise ValueError("Reference threshold (4th col) in the tools config "
                                     "file must be a number at line:\n{}".format(line))

                data[_fields[0]].extend([_fields[2], _ref_thresh])

                if len(_fields) > 4:
                    assert _fields[4] in AVAILABLE_SCOPES, "\'{}\' is not in the list of " \
                                                           "available scopes: {} at line:\n{}".format(_fields[4],
                                                                                                      AVAILABLE_SCOPES,
                                                                                                      line)
                    data[_fields[0]].append(_fields[4])

                if len(_fields) > 5:
                    assert _fields[5] in AVAILABLE_FUNCTIONS, "\'{}\' is not in the list of " \
                                                              "available functions: {} at " \
                                                              "line:\n{}".format(_fields[5], AVAILABLE_FUNCTIONS,
                                                                                 line)
                    data[_fields[0]].append(_fields[5])
    
        if len(data) < 2 and self.is_interrogate_mode is False:
            raise ValueError('Config file requires at least two tools to be compared.')

        return data



import os
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from preprocessing.osutils import ensure_folder_exists
import seaborn as sns

plt.switch_backend('agg')
cmap = sns.diverging_palette(220, 10, as_cmap=True)
from plots.plots_utils import *
from predictions.filters import filter_intronic_bins


def plot_general_bin_info(_df: pd.DataFrame, 
                          outdir: str, 
                          variant_class: str,
                          aggregate_classes: bool,
                          af_column: str):
    """
    Barplot with variant counts on each
    intronic bin
    
    :param pd.DataFrame _df: Input df
    :param str outdir: Output directory
    :param str variant_class: Variant class
    :oaram bool aggregate_classes: Whether splice site are analyzed together with splice region variants
    :param str af_column: Allele frequency column
    """
    
    def _generate_plot(df, filt_to_exclude):
        if df[df['label'].isin([True])].empty or df[df['label'].isin([False])].empty:
            dic = {}

            for _bin in filter_intronic_bins:
                if _bin[0] not in filt_to_exclude:
                    dic[_bin[0]] = np.sum(df.intron_bin == _bin[0])

            plt.bar(dic.keys(), dic.values(), color="silver")

        else:
            ax = sns.countplot(x="intron_bin",
                            hue="outcome",
                            order=[x[0] for x in filter_intronic_bins if
                                    x[0] not in filt_to_exclude],
                            data=df,
                            linewidth=1,
                            edgecolor='k',
                            palette={"Benign": "skyblue",
                                        "Pathogenic": "chocolate"
                                        }
                            )
            for container in ax.containers:
                ax.bar_label(container)
            ax.get_legend().set_title('')

        plt.xlabel("Intron base-pair bins")
        plt.ylabel("Variant counts")
        
    outdir = os.path.join(outdir, "variant_counts")
    ensure_folder_exists(outdir)
    
    fname = os.path.join(outdir, 'counts_per_intronic_bin_{}'.format(variant_class))
    df = _df.copy()
    df.groupby(['intron_bin', 'outcome']).size().to_csv(fname + '.tsv',
                                                        sep="\t")
    
   
    filt_to_exclude = ["all_intronic", "all_except_1-2", "all_except_1-10"]
    if aggregate_classes:
        agg_bins = ['1-2', '3-10']
        opposite = ['1-10']
    else:
        agg_bins = ['1-10']
        opposite = ['1-2', '3-10']
        
    filt_to_exclude.extend(agg_bins)
    
    _generate_plot(df, filt_to_exclude)
    plt.savefig(fname + '.pdf')
    plt.close()

    if af_column in df.columns:
        df[af_column] = pd.to_numeric(df[af_column], downcast='float')

        # df[af_column] = df[af_column].astype('float64')
        # df[af_column].apply(lambda x: '%.10f' % float(x))

        if df[df['label'].isin([True])].empty or df[df['label'].isin([False])].empty:

            ax = sns.scatterplot(x="intron_offset",
                                 y=af_column,
                                 data=df)

        else:
            ax = sns.scatterplot(x="intron_offset",
                                 y=af_column,
                                 data=df,
                                 s=10,
                                 hue="outcome",
                                 palette={"Benign": "skyblue",
                                          "Pathogenic": "chocolate"})
            ax.get_legend().set_title('')

        ax.set(xlabel='Variant intronic offset', ylabel="{} allele frequency".format(af_column))
        plt.tight_layout()
        out = os.path.join(outdir, 'AF_distribution.pdf')
        plt.savefig(out)
        plt.close()


def plot_donor_vs_acceptor(df: pd.DataFrame, 
                           metric: str,
                           fname: str,
                           title: str):

    """
    Plot performance of individual tools on intronic variants
    associated with splicing donor vs splicing acceptors

    :param pd.DataFrame data: Df with stats for each tool
    :pararm str metric: Metric to rank tools
    :param str fname: Output file
    :param str title: Title of the plot
    """
    fig, ax = plt.subplots()
    plt.rcParams.update({'font.size': 10}) 
   
    if len(df['Splice site'].unique().tolist()) == 1:
        df = df.sort_values(metric, ascending=False)
        g = sns.catplot(data=df, 
                        kind="bar",
                        x=metric, 
                        y="Tool",
                        hue="Splice site",
                        palette={'Donor': 'darkred', 'Acceptor': 'skyblue'}, 
                        edgecolor='k',
                        legend_out=True,
                        alpha=.8, 
                        aspect=.8,
                        height=8)

        g.despine(left=True)

    # If both donors and acceptors    
    else:
        sns.stripplot(x=metric, y="Tool", hue='Splice site', linewidth=1, size=10, data=df)
        sns.boxplot(x=metric, y="Tool", data=df, color='white')
   
    plt.legend(loc="upper right",
               bbox_to_anchor=(1.1, 1),
               borderaxespad=0,
               prop=dict(size=8))
    plt.title(title) 
    plt.ylabel('')
    plt.xlim(0, 1)
    plt.tight_layout()
    plt.savefig(fname, bbox_inches='tight')
    plt.close()

def plot_metrics_by_bin_split_ss(df: pd.DataFrame, 
                        fname: str, 
                        aggregate_classes: bool = False):
    """
    :param pd.DataFrame df: Df with a list of
        metrics per each intronic bin
    :param str fname: Output basename
    :param bool aggregate_classes: Whether VETA run is meant to 
    aggregate classes into higher level concepts.
    """

    bins_with_scores = df.bin.unique().tolist()
    bins_to_exclude = ['1-2', '3-10'] if aggregate_classes else ['1-10']
    bins_to_exclude.extend(["all_intronic", "all_except_1-2", "all_except_1-10"])

    df['fraction_scored'] = 1 - df.fraction_nan
    metrics = {"F1": "F1_Score",
               "weighted_F1": "F1 score (weighted)",
               "fraction_scored": "Fraction scored"}

    avg = df.groupby(['tool', 'variant_class']).agg({'F1':'mean',
                             'weighted_F1': 'mean',
                             'fraction_scored': 'mean'})
    
    avg = avg.round(2)
    avg.columns = ["avg_{}".format(x) for x in avg.columns]

    df = pd.merge(df, avg, on=["tool", "variant_class"], how='left')

    n_tools = df.tool.unique().size 
    if n_tools > 10:
        pal = sns.diverging_palette(250, 30, l=65, n=n_tools, center="dark")
        fontsize="small"
    else:
        pal = sns.cubehelix_palette(start=.5, rot=-.5, as_cmap=False, reverse=True, n_colors=n_tools)
        fontsize="medium"
    
    categories=["501-1000", "201-500", "41-200", "11-40"]
    categories.append("1-10") if aggregate_classes else categories.extend(["3-10", "1-2"])
   
    hue_name = 'tool'
    legend_out = False
    for metric, description in metrics.items():
        _df = df.copy()
        d = df[df.variant_class.str.contains('donor')].copy()
        a = df[df.variant_class.str.contains('acceptor')].copy()
        a.bin = pd.Categorical(a.bin, 
                      categories=categories,
                      ordered=True)
        a['tool_with_metric'] = a.tool + " (avg_acc=" + a["avg_{}".format(metric)].astype(str) + ")"
        d['tool_with_metric'] = d.tool + " (avg_don=" + d["avg_{}".format(metric)].astype(str) + ")"
        
        _df = pd.concat([a, d])
        _df = _df.replace({'all_donor_related': 'Donor', 'all_acceptor_related': 'Acceptor'}).sort_values('variant_class')
        _df[metric] = pd.to_numeric(_df[metric])

        # Add avg metric per donor/acceptor
        mean_map = {}
        for _, row in _df.iterrows():
            tool = row.tool
            bin = row.bin
            val = row['avg_{}'.format(metric)]

            if tool in mean_map.keys():
                if bin in mean_map[tool].keys():
                    if ";" in  mean_map[tool][bin]:
                        continue
                    mean_map[tool][bin] = mean_map[tool][bin]  + ";{})".format(val)
                else:
                    mean_map[tool].update({bin: " ({}".format(val)})
            else:
                mean_map[tool] = {bin: " ({}".format(val)}
                         
        mean_df = pd.DataFrame.from_dict(mean_map, orient='index').rename_axis('tool').reset_index()
        mean_df = mean_df.melt(id_vars='tool', value_vars=list(mean_df), var_name='bin', value_name='tool_full')
        _df = pd.merge(_df, mean_df, on=['tool', 'bin'], how='left')
        _df['tool_with_metric'] = _df.tool + _df.tool_full
        hue_name = 'tool_with_metric'
        legend_out = True   
        # Comment code above (from # Add avg ..) to just display tool names in the middle of the two facets
        
        _df = _df.sort_values("avg_{}".format(metric), ascending=False)

        g = sns.catplot(x="bin",
            y=metric,
            kind='point',
            col='variant_class',
            col_order=['Acceptor', 'Donor'],
            order=[i[0] for i in filter_intronic_bins if i[0] not in bins_to_exclude and i[0] in bins_with_scores],
            data=_df,
            sharex=False,
            sharey=False,
            linestyles="-", 
            scale=0.75, 
            aspect=1.2,
            palette=pal,
            markers='s',
            fontsize=fontsize,
            legend= legend_out,
            legend_out = legend_out,
            dodge=False,
            hue=hue_name)

        g.axes[0,0].set_xlabel('Intron bin (bp)')
        g.axes[0,1].set_xlabel('Intron bin (bp)')
        g.axes[0,0].set_ylabel(description)
        g.axes[0,1].set_ylabel(description)

        [plt.setp(ax.texts, text="") for ax in g.axes.flat]
        g.set_titles(row_template = '{row_name}', col_template = '{col_name}')
        
        # Invert acceptor axis
        for i, ax in enumerate(g.axes[0]):
           if i == 0:
               ax.invert_xaxis()
        
        if legend_out:
            # Change label name
            g._legend.set_title("Tool (Mean acceptor; Mean donor)")
        else:
            # Increase space between the plots
            plt.subplots_adjust(hspace=0.4, wspace=0.4)   
            # Place legend in between the plots   
            g.fig.get_axes()[0].legend(loc='upper right', bbox_to_anchor=(1.3, 1)) 
         
        g.set(yticks=np.arange(0, 1.05, 0.1))
        out = fname + '_all_split_ss_' + metric + '.pdf'
        plt.savefig(out, bbox_inches="tight")
        plt.close()
        
def plot_metrics_by_bin(df: pd.DataFrame, 
                        fname: str, 
                        aggregate_classes: bool):
    """
    :param pd.DataFrame df: Df with a list of
        metrics per each intronic bin
    :param str fname: Output basename
    :param bool aggregate_classes: Whether VETA run is meant to 
    aggregate classes into higher level concepts.
    """
    bins_with_scores = df.bin.unique().tolist()

    bins_to_exclude = ['1-2', '3-10'] if aggregate_classes else ['1-10']
    bins_to_exclude.extend(["all_intronic", "all_except_1-2", "all_except_1-10"])
    variant_class = df.name
    
    df['fraction_scored'] = 1 - df.fraction_nan
    df['weighted_roc_auc'] = df.apply(lambda x: None if x.roc_auc is None else x.fraction_scored * x.roc_auc, axis = 1)
    df['weighted_ap_score'] = df.apply(lambda x: None if x.ap_score is None else x.fraction_scored * x.ap_score, axis = 1)
    
    # Keep only tools with auc and ap scores in all bins
    na_rows = df.loc[df.isnull()['weighted_roc_auc'] | df.isnull()['weighted_ap_score']]
    tools_to_remove = na_rows.tool.unique()
    
    _df = df[~df.tool.isin(tools_to_remove)]
    _avg = _df.groupby('tool').agg({'weighted_roc_auc': 'mean',
                                   'weighted_ap_score': 'mean'})
    _avg = _avg.rename(columns={'weighted_roc_auc': 'avg_weighted_roc_auc_all_bins',
                                'weighted_ap_score': 'avg_weighted_ap_score_all_bins'})
    
    metrics = {"F1": "F1_Score",
               "weighted_F1": "F1 score (weighted)",
               "fraction_scored": "Fraction scored",
               "roc_auc": "auROC",
               "weighted_roc_auc": "auROC (weighted)",
               "weighted_roc_auc_all_bins": "auROC (weighted)",
               "ap_score": "Average precision",
               "weighted_ap_score": "Average precision (weighted)",
               "weighted_ap_score_all_bins": "Average precision (weighted)"}
    
    avg = df.groupby('tool').agg({'F1':'mean',
                             'weighted_F1': 'mean',
                             'fraction_scored': 'mean',
                             'roc_auc': 'mean',
                             'ap_score': 'mean',
                             'weighted_roc_auc': 'mean',
                             'weighted_ap_score': 'mean'})
    
    avg = avg.round(3)
    avg.columns = ["avg_{}".format(x) for x in avg.columns]
    df = pd.merge(df, avg, on="tool", how='left').merge(_avg, on='tool', how='left')

    n_tools = df.tool.unique().size 
    if n_tools > 10:
        pal = sns.diverging_palette(250, 30, l=65, n=n_tools, center="dark")
        fontsize="small"
    else:
        pal = sns.cubehelix_palette(start=.5, rot=-.5, as_cmap=False, reverse=True, n_colors=n_tools)
        fontsize="medium"

    n_tools = df.tool.unique().size 

    for _metric, description in metrics.items():
        
        _df = df.copy()
        if "_all_bins" in _metric:
            _df = _df[~df['avg_{}'.format(_metric)].isnull()]   
        metric = _metric.replace("_all_bins", "") 
            
        _df[metric] = pd.to_numeric(_df[metric])
        _df['tool'] = _df.tool + " (mean=" + round(df["avg_{}".format(_metric)], 3).astype(str) + ")"
        
        _df = _df.sort_values("avg_{}".format(_metric), ascending=False)
        g = sns.catplot(x="bin",
            y=metric,
            kind='point',
            order=[i[0] for i in filter_intronic_bins if i[0] not in bins_to_exclude and i[0] in bins_with_scores],
            data=_df,
            linestyles="-", 
            scale=0.75, 
            aspect=1.5,
            palette=pal,
            markers='s',
            fontsize=fontsize,
            legend=False,
            dodge=True,
            hue="tool")

        plt.subplots_adjust(right=0.7)
        plt.legend(loc="upper center",
        bbox_to_anchor=(1.15, 1),
        borderaxespad=0.,
        prop=dict(size=7))

        plt.yticks(np.arange(0, 1.05, 0.1))

        plt.xlabel("Intron bin (bp)")
        plt.ylabel(description)
        plt.tight_layout()
        
        if '_all_bins' in _metric:
            out = fname + '_' + variant_class + "_" + metric + '_tools_predicting_all_bins.pdf'    
        else:
            out = fname + '_' + variant_class + "_" + metric + '.pdf'

        plt.savefig(out, bbox_inches="tight")
        plt.close()
    


import os
from typing import List
import logging
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import seaborn as sns
from matplotlib.lines import Line2D


def plot_area(df: pd.DataFrame, outdir: str):
    """
    Area plot that gives an overview of the
    tools predictions for a given set of
    input variants
    :param pd.DataFrame df: Input df with
    ratios of tools that predict a given
    outcome.
    :param str outdir: Output directory to
        save the fig
    """

    fig, ax = plt.subplots(figsize=(12, 12))

    map_colors = {"Unpredictable": "grey",
                  np.nan: "grey",
                  "Is benign": "darkblue",
                  "Is pathogenic": "darkred"}
    
    ax = df.sort_values(['Is pathogenic', 'Is benign'], ascending=False).plot.area(color=[map_colors.get(x)
                      for x in df.columns], alpha=0.65)
    plt.legend(loc='center left', bbox_to_anchor=(1.0, 0.5))
    ax.set_ylabel('Fraction of tools')
    ax.set_xlabel('Variants (N={})'.format(df.shape[0]))
    plt.xticks([])
    plt.tight_layout()
    plt.savefig(os.path.join(outdir, 'overall_predicted_ratios.pdf'))
    plt.close()


def plot_heatmap(df: pd.DataFrame,
                 outdir: str,
                 display_annot: bool = False,
                 benign_too: pd.DataFrame = None):

    """

    Draw heatmaps that color the ratio of tools
    that predict pathogenicity or unpredictability.
    Each row is a variant, the more tools predict
    a given outcome, the redder the color

    :param pd.DataFrame df: Input df with
    ratios of tools that predict a given
    outcome.
    :param str outdir: Output directory to
        save the fig
    :param bool display_annot: Display row
        annotations in the heatmap (variant ID).
        Default: `False`. Useful when few
        variants are being plotted.
    :param pd.DataFrame benign_too: Display 
    also variants for which most tools predict 
    to be benign. Default: None, display
    unpredictable variants
    :return:
    """

    plot_benign = False
    fig, ax = plt.subplots(1, 2, figsize=(5, 7))

    if display_annot:

        if isinstance(benign_too, pd.DataFrame):
            if benign_too.shape[0] > 0 and df.shape[0] > 0:
                all_path = df[df['Is pathogenic'] == 1].copy()
                if all_path.shape[0] > 50:
                    df = all_path
                else:
                    df = df.sort_values(['Is pathogenic']).head(50)
                benign_too = benign_too.sort_values(['Is benign']).head(50)
                outfile = os.path.join(
                    outdir, 'predictability_trade_off_top_patho_and_benign.pdf')
                plot_benign = True
            else:
                plt.close()
                return

        elif df.shape[0] > 0:
            df = df.sort_values(['Is pathogenic']).head(50)
            outfile = os.path.join(
                outdir, 'predictability_trade_off_top_candidates.pdf')
        else:
            plt.close()
            return
    else:
        outfile = os.path.join(outdir, 'predictability_trade_off.pdf')

    if plot_benign:

        sns.heatmap(df[["Is pathogenic"]][::-1], ax=ax[0],
                    cbar=False,
                    vmax=1,
                    vmin=0,
                    cmap="OrRd",
                    linewidths=.5,
                    linecolor='k',
                    yticklabels=display_annot)

        sns.heatmap(benign_too[["Is benign"]][::-1], ax=ax[1],
                    cbar_kws={'label': 'Fraction of tools'},
                    vmax=1,
                    vmin=0,
                    cmap="OrRd",
                    linewidths=.5,
                    linecolor='k',
                    annot_kws={'size': 1},
                    yticklabels=display_annot)

        ax[0].set_yticklabels(ax[0].get_ymajorticklabels(), fontsize=10)
        ax[1].set_yticklabels(ax[1].get_ymajorticklabels(), fontsize=10)

        ax[0].set_ylabel('')
        ax[1].set_ylabel('')

    else:
        _linewidth = 0.001 if df.shape[0] > 2000 else .5
        sns.heatmap(df[["Is pathogenic"]][::-1], ax=ax[0], cbar_kws={'label': 'Fraction of tools'},
                    vmax=1,
                    vmin=0,
                    cmap="OrRd",
                    linewidths=_linewidth,
                    linecolor='k',
                    yticklabels=display_annot)

        sns.heatmap(df[["Unpredictable"]], ax=ax[1], cbar_kws={'label': 'Percentage (%)'},
                    vmax=100,
                    vmin=0,
                    cmap="OrRd",
                    linewidths=_linewidth,
                    linecolor='k',
                    yticklabels=False)
        ax[0].set_ylabel('All variants ({})'.format(df.shape[0]))
        ax[1].set_ylabel('')


    plt.tight_layout()
    plt.savefig(outfile)
    plt.close()


def plot_heatmap_toptools(df: pd.DataFrame, filters, outdir):
    """
    Draw heatmap for the top tools

   :param pd.DataFrame df: Df with predictions
        for each variant
    :param List filters: Location filters to employ
        so that variants at each location are processed
        independently
    :param str outdir: Output directory
    """
    from plots.plots_benchmark_mode import plot_heatmap

    if 'label' in df.columns:
        df.rename(columns={"label": "Ground Truth (*)"}, inplace=True)
    
    os.makedirs(os.path.join(outdir, 'out_heatmaps'), exist_ok=True)
    outdir = os.path.join(outdir, 'out_heatmaps')

    for filter_name in filters:
        
        df_f = df[df.location == filter_name].copy()
        df_f = df_f.drop(columns=['variant_id', 'HGVSc', 'HGVSg', 'location', "SYMBOL"])
        if df_f.shape[0] < 5:
            continue

        fname = os.path.join(
            outdir, "top_tools_heatmap_{}.pdf".format(filter_name))

        try:
            plot_heatmap(df_f, fname,
            cluster_rows=True, 
            skip_preparation=True)
        except RecursionError:
            logging.info("WARN: Error when generating heatmap for '{}' variants".format(filter_name))
            pass


def plot_accuracy(stats_df: pd.DataFrame,
                  metric: str,
                  location: str,
                  out_dir: str):
    """
    Plots weighted_accuracy bars

    :param pd.DataFrame stats_df: Stats dataframe
    :param str metric: Metric to rank
    :param str location: Variants location
    :param str out_dir: Output directory
    """
    stats_df = stats_df.sort_values([metric])
    n_tools = len(stats_df)
    if n_tools < 5:
        figsize = (3, 3)
    elif 5 <= n_tools <= 30:
        figsize = (5, 5)
    else:
        figsize = (7, 7)

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.barh(range(stats_df.shape[0]), stats_df[metric],
            color='darkgrey',
            edgecolor='k',
            linewidth=1)

    plt.axvline(0.9, color='b', linestyle='--')
    xlabel = (metric[:1].upper() + metric[1:]).replace("_", " ")
    _title = (location[:1].upper() + location[1:]).replace("_", " ")
    plt.xlabel(xlabel)
    plt.title("{} variants (N={})".format(
        _title, str(stats_df.iloc[0, :].total)))

    plt.xlim(left=0)
    plt.yticks(range(stats_df.shape[0]), stats_df['tool'])
    plt.ylim(-1, stats_df.shape[0])
    fig.tight_layout()
    plt.savefig(os.path.join(out_dir, "performance_{}.pdf".format(location)))
    plt.close()


def plot_tool_score_distribution(_df: pd.DataFrame,
                                 tool: str,
                                 thresholds: List,
                                 outdir: str):
    """
    Plot score distribution for a given tool
    :param pd.DataFrame _df: Input df
    :param str tool: Specific tool to analyse
    :param List thresholds: List of tools with the
        reference thresholds
    :param outdir: Output directory
    """
    os.makedirs(os.path.join(outdir, 'out_score_distribution'), exist_ok=True)
    outdir = os.path.join(outdir, 'out_score_distribution')
    plt.subplots(figsize=(7, 5))
    nas = sum(pd.isnull(_df[tool]))

    df = _df.copy()
    df[tool].dropna(inplace=True)
    p = sns.histplot(df[tool], bins=30, color='slateblue', edgecolor='black')
    p.tick_params(labelsize=13)
    plt.xlabel('{} ({} % missing data)'.format(
        tool, round(nas / _df.shape[0] * 100, 3)), fontsize=13)

    p.yaxis.get_label().set_fontsize(13)
    plt.axvline([x[2] for x in thresholds if x[0] == tool],
                color='r',
                linestyle="--")

    legend_element = [Line2D([0], [0], color='r', lw=4,
                             label='Reference threshold')]

    plt.legend(handles=legend_element, loc='best')
    plt.savefig(os.path.join(outdir, "score_dist_{}.pdf".format(tool)))
    plt.close()



import fnmatch
from collections import defaultdict
from typing import List
import pandas as pd

from base import Base
from plots.plots_benchmark_mode import *
from plots.plots_machine_learning import plot_feature_correlation
from predictions import introns
from predictions import metrics
from predictions.apply import apply_tool_predictions
from predictions.metapredictions import Metapredictions
from predictions.thresholds import perform_threshold_analysis
from preprocessing.clinvar import *
from preprocessing.tables import generate_clinvar_table, generate_consequence_table
from preprocessing.osutils import ensure_folder_exists


class BenchmarkTools(Base):
    """
    Benchmark mode to compare tools
    based on a reference dataset
    """

    def __init__(self, dataset: str,
                 out_dir: str,
                 scopes_to_predict: List = None,
                 types_of_variant: List = None,
                 metric: str = "weighted_accuracy",
                 location: str = "HGVSc",
                 aggregate_classes: bool = False,
                 select_conseqs: str = "gene_body",
                 do_intronic_analysis: bool = False,
                 split_splice_sites: bool = False,
                 clinvar_stars: str = "1s_l",
                 phenotype_ids: list = None,
                 do_threshold_analysis: bool = False,
                 do_bootstrapping: bool = False,
                 do_machine_learning: bool = False,
                 allele_frequency_col: str = "gnomADg_AF",
                 skip_heatmap: bool = False,
                 tools_config: str = None):

        """
        ----
        Base args described in Base class
        ----
        """

        dataset, is_clinvar = self.check_dataset_arg(dataset)
        super().__init__(vcf=dataset,
                         out_dir=out_dir,
                         scopes_to_predict=scopes_to_predict,
                         types_of_variant=types_of_variant,
                         metric=metric,
                         location=location,
                         aggregate_classes=aggregate_classes,
                         select_conseqs= select_conseqs,
                         do_intronic_analysis=do_intronic_analysis,
                         split_splice_sites=split_splice_sites,
                         is_clinvar=is_clinvar,
                         allele_frequency_col=allele_frequency_col,
                         skip_heatmap=skip_heatmap,
                         tools_config=tools_config)

        self.clinvar_stars = clinvar_stars
        self.do_threshold_analysis = do_threshold_analysis
        self.do_bootstrapping = do_bootstrapping
        self.do_machine_learning = do_machine_learning

        # Remove tools where all values are missing
        tools = [t[0] for t in self.thresholds]
        self.df.dropna(axis=1, how="all", inplace=True)
        self.null_tools = list(set(tools) - set(list(self.df)))
        if len(self.null_tools) > 0:
            logging.info("Tools with no predictions were found. They will be "
                         "discarded from analysis: {}".format(self.null_tools))

        if self.is_clinvar:
            self.out_dir = os.path.join(self.out_dir, self.clinvar_stars)

            self.df = filter_by_condition(self.df, phenotype_ids)
            
            df_clinvar_sure = filter_clinvar_sure(self.df)
            self.clinvar_levels_dict = {
                '0s': df_clinvar_sure,
                '1s': filter_clinvar_1_star(df_clinvar_sure),
                '2s': filter_clinvar_2_stars(df_clinvar_sure),
                '3s': filter_clinvar_3_stars(df_clinvar_sure),
                '4s': filter_clinvar_4_stars(df_clinvar_sure),
                '0s_l': self.df,
                '1s_l': filter_clinvar_1_star(self.df),
                '2s_l': filter_clinvar_2_stars(self.df),
                '3s_l': filter_clinvar_3_stars(self.df),
                '4s_l': filter_clinvar_4_stars(self.df)}

            generate_clinvar_table(self.clinvar_levels_dict, 
                                   self.variant_types, 
                                   self.out_dir,
                                   self.clinvar_stars)

            self.df = self.clinvar_levels_dict[self.clinvar_stars]
            assert self.df.shape[0] > 0, "No remaining variants after filtering by {} " \
                                         "level. Try a more relaxed value for the '-c'" \
                                         " arg.".format(self.clinvar_stars)
            logging.info('Number of variants after filtering by Clinvar stars ({}): {}'.format(self.clinvar_stars, self.df.shape[0]))

        generate_consequence_table(self.df, 
                                   self.out_dir)

        self.top_tools, f1_at_ref_threshold = self.do_performance_comparison()
        if self.do_intronic_analysis:
            thresholds = [tool for tool in self.thresholds if tool[3] != 'Protein']
         
            introns.IntronicAnalysis(self.df,
                                     self.split_splice_sites,
                                     thresholds,
                                     self.metric,
                                     self.aggregate_classes,
                                     self.out_dir,
                                     self.allele_frequency_col)

        if self.do_threshold_analysis:
            
            new_thresholds = perform_threshold_analysis(self.df,
                                                        self.location_filters,
                                                        self.thresholds,
                                                        self.tools_config,
                                                        self.out_dir,
                                                        f1_at_ref_threshold,
                                                        do_bootstrapping=self.do_bootstrapping)

        if self.do_machine_learning:
            self.do_ml_analysis()

    def do_performance_comparison(self):
        """
        Generates performance stats and plots
        for the processed df

        :return dict: df with predictions
        """
        logging.info("----------------------------------")
        logging.info("Tools performance analysis started")
        logging.info("----------------------------------")

        top_tools, f1_at_ref_threshold = {}, {}
        if self.is_clinvar:
            logging.info("Evaluations are done with Clinvar variants that belong "
                         "to the filtering strategy employed ({}). Playing with '--clinvar_stars' "
                         "argument allows using more permissive or restrictive variant sets.".format(self.clinvar_stars))
            
            
        for var_type, _var_type_func in self.variant_types:
            outdir = os.path.join(self.out_dir, "tools_benchmark", var_type)
            ensure_folder_exists(outdir)

            _df_v = _var_type_func(self.df).copy()
            print()
            logging.info("----------")
            logging.info("Looking at {} ({} variants)".format(var_type, _df_v.shape[0]))
            logging.info("----------")
            print()
            
            if _df_v.shape[0] == 0:
                logging.warning("No variants. Skipping!".format(var_type))
                continue
            
            for _location in self.location_filters:
                
                if _location == "all":
                    df = _df_v.copy()
                else:
                    df = _df_v[_df_v.location == _location].copy()
                    
                if df.empty:
                    continue
                
                statistics = defaultdict(list)
                        
                n_pos = df.label.sum()
                n_neg = (~ df.label).sum()
                df['count_class'] = df.groupby('outcome')['outcome'].transform('size')
                
                logging.info("----------")
                logging.info("{} variants (n={})".format(_location, df.shape[0]))
                logging.info("Pathogenic: {}".format(n_pos))
                logging.info("Benign: {}".format(n_neg))
                logging.info("----------")
                print()
                
                if df.shape[0] < 10:
                    logging.info("Less than 10 '{}' '{}' variants found ({})"
                                ". Skipping!".format(var_type, _location, df.shape[0]))
                    continue
                
                df = self.apply_thresholds(df, _location)

                if not self.skip_heatmap:
                    if df.shape[0] > 10:
                        logging.info("Generating performance heatmaps for '{}' '{}' variants. "
                                    "If dataset is large, this step will take long time.".format(var_type, _location))
                        
                        fname = os.path.join(outdir, "performance_at_fixed_thresh", "heatmap_" + _location + '.pdf')
                        plot_heatmap(df, fname=fname)
                        logging.info("Done")
                    else:
                        logging.info("Less than 10 variants located in '{}' of type '{}'. " \
                                    "Skipping heatmap generation.".format(_location, var_type))

                ensure_folder_exists(os.path.join(outdir, "results_tsv"))
                out_preds = os.path.join(outdir, "results_tsv", "preds_{}_{}_class.tsv").format(var_type, _location)
                out_preds_scores = os.path.join(outdir, "results_tsv", "preds_{}_{}_scores.tsv").format(var_type, _location)
                out_c = ['chr', 'pos', 'ref', 'alt', 'SYMBOL'] + [x for x in list(df) if '_prediction' in x] + ['label']
                out_c_scores = ['chr', 'pos', 'ref', 'alt', 'SYMBOL'] + [x for x in list(df) if '_prediction' not in x and x not in ['count_class', 'blank', 'label']]
  
                df[out_c].to_csv(out_preds, index=False, sep="\t")
                df[out_c_scores].to_csv(out_preds_scores, index=False, sep="\t")
                roc_metrics_per_tool, pr_metrics_per_tool = [], []

                for tool, direction, _, _ in self.thresholds:

                    try:

                        statistics = metrics.generate_statistics(df, statistics, _location, tool)
                        scored = df[~df[tool + '_prediction'].isnull()]
                        if scored.empty:
                            logging.info("{} did not score any {} variant.".format(tool, _location))
                            continue
                       
                        if 0 < scored.shape[0] < df.shape[0] * 0.05:
                            logging.info("Warn: {} scored some '{}' variants (scored={}), but they account for less than "
                                        "5% of the dataset size. Plots will not show these results.".format(tool, _location, scored.shape[0]))
                            continue

                        # Distributions of each class are only plotted for SNPs and
                        # all types for splice_site and coding locations
                        if (_location in ['all', 'splice_site', 'splice_region', 'intronic', 'missense', 'coding'] and var_type in ['snps', 'all_types']):
                            plot_density_by_class(df[[tool, "label"]],
                                                thresholds=self.thresholds,
                                                fname=os.path.join(outdir, 'class_distribution', tool + "_" + _location))

                        ####################
                        ### ROC analysis ###
                        ####################
                        # Required to have predictions on variants of both classes
                        n_pos_pred = scored[scored.label].shape[0]
                        n_neg_pred = scored[scored.label == False].shape[0]

                        if scored.empty or 0 < scored.shape[0] < df.shape[0] * 0.33:
                            logging.info("Warn: {} scored some '{}' variants (scored={}), but they account for less than "
                                        "33% of the dataset size. ROC analysis will be skipped for this tool.".format(tool, _location, scored.shape[0]))
                        
                        elif min(n_pos_pred, n_neg_pred) < 15:
                            logging.info("Warn: No minimum number of predictions on each class (15) found. Skipping ROC analysis for {} tool in {} variant set".format(tool, _location))
                            
                        else:
                    
                            na_frac = 1 - (scored.shape[0] / df.shape[0])

                            try:
                                roc_curve, pr_curve, roc_auc, ap_score = metrics.do_roc_analysis(scored[[tool, 'label']],
                                                                                                tool,
                                                                                                higher_is_better=direction == ">")
            
                                na_frac = 1 - (scored.shape[0] / df.shape[0])
                                roc_metrics_per_tool.append([tool, na_frac, roc_curve[0], roc_curve[1],
                                                            roc_curve[2], roc_curve[3], roc_auc])
                                pr_metrics_per_tool.append([tool, na_frac, pr_curve[0], pr_curve[1],
                                                            pr_curve[2], pr_curve[3], ap_score])
                            except TypeError:

                                roc_metrics_per_tool.append([tool, na_frac, None, None, None, None, None])
                                pr_metrics_per_tool.append([tool, na_frac, None, None, None, None, None])
                                pass
                        
                    except KeyError:
                        pass
                
                stats_all_df = pd.DataFrame(statistics)
    
                if roc_metrics_per_tool:
                    roc_m = pd.DataFrame([[x[0], x[-1]] for x in roc_metrics_per_tool], columns=['tool', 'auROC'])
                    pr_m = pd.DataFrame([[x[0], x[-1]] for x in pr_metrics_per_tool], columns=['tool', 'average_precision'])
                    stats_all_df = stats_all_df.merge(roc_m, on='tool', how='left').merge(pr_m, on='tool', how='left')

                # draw plots
                out_stats = os.path.join(outdir, "results_tsv", "statistics_{}_{}.tsv").format(var_type, _location)
                out_ranks = os.path.join(outdir, "results_tsv", "tools_ranking_{}_{}.tsv").format(var_type, _location)

                af_plot = os.path.join(outdir, "allele_frequency", "AF_{}_{}.pdf".format(var_type, _location))
                unscored_plot = os.path.join(outdir, "performance_at_fixed_thresh", "unscored_fraction_{}_{}.pdf".format(var_type, _location))
                barplot_plot = os.path.join(outdir, "performance_at_fixed_thresh", "barplot_{}_{}.pdf".format(var_type, _location))
                metrics_plot = os.path.join(outdir, "performance_at_fixed_thresh", "scatter_{}_{}.pdf".format(var_type, _location))

                plot_allele_frequency(df, af_plot, self.allele_frequency_col)
                plot_unscored(stats_all_df, unscored_plot)
                plot_tools_barplot(stats_all_df, barplot_plot, self.metric)
                plot_metrics(stats_all_df, metrics_plot, self.metric)

                for i, _roc_data in enumerate([roc_metrics_per_tool, pr_metrics_per_tool]):

                    if i == 0:
                        is_roc = True 
                        bname = "ROC_"
                    else:
                        is_roc = False
                        bname = "ROC_pr_"
                    
                    if _roc_data:
                        
                        _ = plot_curve(_roc_data, 
                                os.path.join(outdir, "roc_analysis", "{}{}.pdf".format(bname, _location)),
                                (n_pos, n_neg),
                                is_roc=is_roc)

                stats_all_df.drop(['filter'], axis=1).to_csv(out_stats, sep="\t", index=False)

                ranked_stats = stats_all_df[["tool", "weighted_accuracy", "accuracy",
                                            "weighted_F1", "F1", "coverage",
                                            "specificity", "sensitivity",
                                            "precision", "mcc", "norm_mcc", "weighted_norm_mcc"]].sort_values([self.metric], ascending=False)

                top_tools[_location] = ranked_stats.head(20) if ranked_stats.shape[0] > 20 else ranked_stats
                f1_at_ref_threshold[_location] = dict(zip(ranked_stats.tool, ranked_stats.F1))
                ranked_stats.to_csv(out_ranks, sep="\t", index=False)

        logging.info("Done!")
        return top_tools, f1_at_ref_threshold

    def apply_thresholds(self, df: pd.DataFrame,
                         filter_location: str,
                         new_thresholds: pd.DataFrame = None):
        """
        Apply tools predictions based on a
        list of thresholds

        :param pd.DataFrame df: Input df
        :param str filter_location: Location filter
        :param pd.DataFrame new_thresholds: Df of new
            thresholds to evaluate predictions.
            Only applicable after a threshold
            analysis was performed by setting
            `--do_threshold_analysis` in VETA.

        :return pd.DataFrame: Df with tools predictions
            based on a list of reference thresholds
        """

        if new_thresholds is not None:
            ntlines = []

            for _t in new_thresholds:
                ntline = [x for x in _t]

                if _t[0] in new_thresholds[filter_location]:
                    ntline[2] = new_thresholds[filter_location][_t[0]]
                ntlines.append(ntline)

            df = apply_tool_predictions(df, ntlines)
            suffix = "_proposed"
            return df, suffix

        else:
            return apply_tool_predictions(df, self.thresholds)

    def check_dataset_arg(self, dataset: str):
        """
        Checks the input dataset argument
        :param str dataset:
        :return str: Dataset info
        :return bool: Whether dataset
            refers to clinvar (if a file)
        """
        # if reference datasets are in directory
        if os.path.isdir(dataset):
            fname_benign = [filename for filename in os.listdir(dataset) if
                            any(fnmatch.fnmatch(filename, pattern) for pattern in ["*benign*[vcf.bgz]",
                                                                                   "*neutral*[vcf.bgz]"])]
            fname_patho = [filename for filename in os.listdir(dataset) if
                           any(fnmatch.fnmatch(filename, pattern) for pattern in ["*deleterious*[vcf.bgz]",
                                                                                  "*pathogenic*[vcf.bgz]"])]
            if fname_benign and fname_patho:
                benign = os.path.join(dataset, fname_benign[0])
                pathogenic = os.path.join(dataset, fname_patho[0])
                return (benign, pathogenic), False
            else:
                raise ValueError('Files with required patterns absent in {}'.format(dataset))

        # if a file, veta expects to be from Clinvar
        elif os.path.isfile(dataset):
            return dataset, True

        else:
            raise ValueError("Set a valid value for the input dataset "
                             "(directory or file)")

    def do_ml_analysis(self):
        """
        Machine learning analysis of tools scores.
        """
        logging.info("---------------------------------")
        logging.info("Starting meta-prediction analysis")
        logging.info("---------------------------------")
        out_dir = os.path.join(self.out_dir, "machine_learning")
        ensure_folder_exists(out_dir)

        for _loc in self.location_filters:
            print()
            logging.info("----------")
            logging.info("Analyzing '{}' variants.".format(_loc))
            logging.info("----------")
            if _loc == "all":
                df_f = self.df.copy()
            else:
                df_f = self.df[self.df.location == _loc].copy()

            if df_f.shape[0] < 100:
                logging.info("Less than 100 variants in '{}' dataframe (N={}). "
                             "Machine learning analysis will be skipped.".format(_loc, df_f.shape[0]))
                continue
            
            n_pos_pred = df_f[df_f.label].shape[0]
            n_neg_pred = df_f[df_f.label == False].shape[0]
            if min(n_pos_pred, n_neg_pred) < 30:
                logging.info("Less than 30 variants of the minority class (only {}) in '{}' "
                             "variants. Classifiers will not be trained.".format(min(n_pos_pred, n_neg_pred), _loc))
                continue
            
            ml_data = Metapredictions(df_f, _loc, self.thresholds,
                                      out_dir,
                                      top_tools=self.top_tools,
                                      rank_metric=self.metric)

            plot_feature_correlation(ml_data.df, _loc, out_dir)

            if len(ml_data.features) < 3:
                logging.info("Less than 3 features (aka tool scores) available"
                             " ({}) after removing predictors with many missing "
                             "values in '{}' variants. Classifiers will not be "
                             "trained for such small array size.".format(len(ml_data.features),
                                                                         _loc))
                continue

            logging.info("Applying classifiers to data.")
            trained_clfs = ml_data.do_classification()

            logging.info("Generating feature ranking.")
            ml_data.generate_feature_ranking(trained_clfs)
            logging.info("Generating decision tree plot")
            ml_data.generate_tree_plot(trained_clfs)



import pandas as pd
import seaborn as sns
from dtreeviz.trees import *
from sklearn.feature_selection import _rfe
from predictions.utils_classifiers import *


def plot_feature_correlation(df: pd.DataFrame, 
                             location: str, 
                             outdir: str,
                             drop_AF: bool = True):
    """
    Generates Pearson's feature correlation

    :param pd.DataFrame df: Input dataframe
    :param str location: Locations of the variants
    :param str outdir: Output directory
    """
    cmap = sns.diverging_palette(220, 10, as_cmap=True)

    if "gnomADg_AF" in df.columns and drop_AF:
        df.drop('gnomADg_AF', axis=1, inplace=True)

    if 'intron_offset' in df.columns:
        corr = df.drop(['label', 'intron_offset'], axis=1).corr()
    else:
        corr = df.drop('label', axis=1).corr()

    # Generate a mask for the upper triangle
    mask = np.zeros_like(corr, dtype=np.bool)
    mask[np.triu_indices_from(mask)] = True

    annot=False if len(corr.columns) > 15 else True
    f, ax = plt.subplots(figsize=(11, 9))
    ax = sns.heatmap(corr,
                     mask=mask,
                     cmap=cmap,
                     vmax=1,
                     vmin=-1,
                     annot=annot, 
                     square=True,
                     linewidths=.5,
                     linecolor='k',
                     cbar_kws={'shrink': 0.5},
                     ax=ax)

    ax.axhline(y=0, color='k', linewidth=5)
    ax.axhline(y=corr.shape[1], color='k', linewidth=5)
    ax.axvline(x=0, color='k', linewidth=5)
    ax.axvline(x=corr.shape[0], color='k', linewidth=10)
    ax.set_xticklabels(ax.get_xticklabels(), rotation=70)
    plt.tight_layout()
    outdir = os.path.join(outdir, 'feature_correlation')
    os.makedirs(outdir, exist_ok=True)
    plt.savefig(os.path.join(outdir, 'pearson_correlation_{}.pdf'.format(location)))
    plt.close()


def plot_feature_importance(importance_scores: dict,
                            method: str,
                            location: str,
                            out_dir: str,
                            std: np.array = None):
    """
    Plot feature importance scores

    :param dict importance_scores: Feature names with corresponding
        importance scores
    :param str method: Whether scores are from random forests
    `feature_importance` attribute or from information gain analysis
    :param str location: Location of the variants
    :param str out_dir: Output directory
    :param np.array std: Standard deviation for the importance scores.
    """
    fig = plt.figure()

    # Plot just top 20
    if len(importance_scores) > 20:
        importance_scores = {k: v for (k, v) in [x for x in importance_scores.items()][-20:]}

        if std is not None:
            std = std[-20:]

    if method == 'feature_importance':

        p = plt.barh(range(len(importance_scores)), importance_scores.values(),
                     xerr=std,
                     error_kw={'markeredgewidth': 10},
                     capsize=1,
                     color='darkgrey',
                     edgecolor='k',
                     linewidth=1)
        plt.xlabel('Feature Importance')

    elif method == 'information_gain':
        out_dir = os.path.join(out_dir, "mutual_information_gain")
        os.makedirs(out_dir, exist_ok=True)
        p = plt.barh(range(len(importance_scores)), importance_scores.values(),
                     color='darkgrey',
                     edgecolor='k',
                     linewidth=1)
        plt.xlabel('Information Gain')

    plt.xlim(left=0)
    plt.yticks(range(0, len(importance_scores)), importance_scores.keys())
    plt.ylim([-1, len(importance_scores)])
    fig.tight_layout()
    plt.savefig(os.path.join(out_dir, "{}_{}.pdf".format(method, location)))
    plt.close()


def plot_rfecv(selector: _rfe.RFECV,
               min_features: int,
               feature_names: list,
               clf_name: str,
               loc: str,
               out_dir: str):
    """
    Plot cross-validation performance after recursive feature elimination

    :param _rfe.RFECV selector: Feature selector
    :param int min_features: Minimum number of features used
    :param list feature_names: Name of the features to plot
    :param str clf_name: Name of the classifier fitted
    :param str loc: Location of the variants
    :param str out_dir: Output directory
    :return:
    """
    selected_idx = [i for i, x in enumerate(selector.support_) if x]
    selected_tools = [feature_names[i] for i in selected_idx]

    fig = plt.figure()

    plt.plot(range(min_features,
                   len(selector.cv_results_['mean_test_score']) + min_features),
             selector.cv_results_['mean_test_score'])

    plt.axvline(selector.n_features_,
                color='red',
                linestyle='--',
                linewidth=0.5)

    plt.gcf().text(0.75, 0.5,
                   'Selected features:\n{}'.format('\n'.join(selected_tools)),
                   fontsize=8)

    plt.xlabel("Number of features selected")
    plt.ylabel("Cross Validation score")
    plt.title("Number of features for best score: {}".format(selector.n_features_))
    plt.ylim(0.5, 1)
    plt.subplots_adjust(left=0.3)
    fig.tight_layout()
    plt.savefig(os.path.join(out_dir, "rfecv_{}_{}.pdf".format(loc, clf_name)))
    plt.close()


def plot_decision_tree(dt_pipeline: Pipeline,
                       x: np.array, y: np.array,
                       feature_names: list,
                       loc: str,
                       out_dir: str):
    """
    Generate decision tree decisions with dtreeviz packages
    :param dt_pipeline: Fitted estimator
    :param np.array x: Instances to run through the model
    :param np.array y: Labels
    :param list feature_names: Names of the features
    :param str loc: Location
    :param str out_dir: Output directory
    """
    _clf = dt_pipeline.named_steps['dt']
    out_dir = os.path.join(out_dir, "decision_tree")
    os.makedirs(out_dir, exist_ok=True)
    for bool_ in [False, True]:
        viz = dtreeviz(_clf,
                       x,
                       y,
                       target_name='Pathogenicity',
                       feature_names=feature_names,
                       fancy=bool_,
                       class_names=['Benign', 'Pathogenic', 'unknown'])

        out_str = "tree_{}_fancy.svg".format(loc) if bool_ else "tree_{}.svg".format(loc)
        viz.save(os.path.join(out_dir, out_str))



import logging
from typing import Union
import numpy as np
import pandas as pd


def ratio(a, b):
    if b == 0:
        return 0
    return float(a) / float(b)


def tuple2float(x: tuple):
    """
    Returns predictions for a given
    tool.

    :param tupple x: Tuple
    with info about the tool (1st elem)
    and its predictions (2nd elem).

    :return list: List of predictions (2nd elem)
    """

    return x[1] if x[1] != '.' or x[1] != '' or x[1] else np.nan


# def custom_range(series, steps):
#     mi = series.min()
#     ma = series.max()
#     step = (ma - mi) / float(steps)
#     return np.arange(mi, ma, step)


def to_numeric(preds: list, absolute: bool = True):
    """
    Converts a prediction score
    to a numeric value. This is
    for tools that output a single 
    numeric value 
    :param list preds: List with 
    predictions for a single variant.
    List is as long as the number 
    of VCF annotations that compose 
    the tool.

    :param bool absolute: Whether absolute value
    should be returned (e.g. splicing tools that
    predict PSI changes may harbor a negative
    magnitude value. Default: `True`.
    :return float: Single numeric prediction
    """

    # by default, we expect a single
    # VCF annotation for such tools
    # (e.g. CADD)
    if len(preds) > 1:
        raise ValueError('Multiple VCF fields were used in the '
                         'config file in a tool that expects a '
                         'single field with a single numeric value. '
                         'Error: {}'.format(preds))

    _p = preds[0]
    if isinstance(_p, float):
        score = round(_p, 4)

    else:
        try:
            if any([k in _p for k in [',', '|', '&']]):
                raise ValueError('Multiple predictions were found '
                                 'for a tool that expects a single '
                                 'numeric value: {}'.format(_p))

            score = round(float(_p), 4)

        except (TypeError, ValueError):
            return np.nan

    return abs(score) if absolute else score


def categorical_to_numerical(pred: list):
    """
    Parse single prediction that comes as 
    a category and transform it in a numeric
    value. It expects the categorical prediction
    to represent classification labels as seen
    in clinvar. Tools such as cVEP and EVE use
    this taxonomy. 
    
    :param list pred: Single prediction. 
    """

    class_map = {'benign': 0,
                 'likely_benign': 0.25,
                 'uncertain_significance': None,
                 'uncertain': None,
                 'conflicting': None,
                 'likely_pathogenic': 0.75,
                 'pathogenic': 1
                 }
    
    if len(pred) == 1:
        pred = pred[0]

        if pred is None:
            return np.nan

        pred = pred.lower()
        all_p = pred.split(",")
    
        # If multiple predictions
        if len(all_p) > 1:

            is_benign = all(["benign" in x for x in all_p])
            is_patho = all(["pathogenic" in x for x in all_p])
            
            if is_benign and is_patho:
                pred = "conflicting"
            
            else:
                pred = all_p[0]
        
        return class_map[pred]
    else:
        raise ValueError('Unexpected number of different preds found: {}'.format(pred))
        
    
def parse_vep_or_vcfanno(pred: str, is_max: bool,
                         _val: float,
                         absolute: bool,
                         average: bool):
    """
    Parse single prediction that harbors
    multiple values (e.g. VEP uses '&' as
    separator, vcnanno used ',') and returns
    an updated _max

    :param str pred: String with multiple scores
    to extract maximum value from
    :param bool is_max: Whether max value should
    be returned. Default: `True`. If `False`,
    the objective is to return the minimum value

    :param float _val: Top score so far
    :param bool absolute: Whether absolute values
    should be used to calculate max score
    :param bool average: Whether values should
    be average. If `True` average is calculated
    on the real values and only at the end the
    absolute number is calculated, if `absolute`
    is `True`

    :return float: Updated `_val` score
    """
    sep = "," if "," in pred else "&"
    _avg = []

    for s in pred.split(sep):
        if not s or s == 'None' or s == ".":
            continue
        s = abs(float(s)) if absolute else float(s)
        if average:
            _avg.append(float(s))
        else:
            s = abs(float(s)) if absolute else float(s)
            if np.isnan(_val):
                _val = s

            elif (is_max and s > _val) or (not is_max and s < _val):
                _val = s

    if average:
        s = abs(np.mean(_avg)) if absolute else np.mean(_avg)
        if (is_max and s > _val) or (not is_max and s < _val):
            _val = s
    return _val


def get_top_pred(preds: list,
                 is_max: bool = True,
                 absolute: bool = True,
                 average: bool = False):
    """
    Return the top scored predicion based
    on the maximum value found for a
    single tool. It may may be composed by
    multiple scores. Example of such tool
    is 'dbscSNV', which has two scores, 
    the max of them will be returned. 
    CAPICE (that has SNV or Indel scores)
    may also fit for this function.

    :param list preds: List of Predictions
    where each prediction contains a single
    score (or multiple scores separated by
    '&'. ',' or '|')

    :param bool is_max: Whether max value should
    be returned. Default: `True`. If `False`,
    the objective is to return the minimum value

    :param bool absolute: Whether absolute value
    should be returned. Maximum value will consider
    the absolute magnitude over the possible
    multiple predictions (including negative values,
    e.g. dPSI predictors). Default: `True`. If `False`
    maximum positive value will be returned

    :param bool average: Whether values of a single
    prediction (when multiple scores exist) should be
    averaged. If `True`, max between averaged multiple
    predictions will be returned.

    :return float: Top prediction
    """
    if all(not v for v in preds):
        return np.nan

    _top = np.nan
    for _p in preds:
        if _p and np.isnan(_top):
            try:
                _top = abs(float(_p)) if absolute else float(_p)

            # if multiple
            except ValueError:
                _top = parse_vep_or_vcfanno(_p,
                                            is_max=is_max,
                                            _val=_top,
                                            absolute=absolute,
                                            average=average)

        elif _p:
            try:
                if (is_max and float(_p) > _top) or (not is_max and float(_p) < _top):
                    _top = float(_p)

            except ValueError:
                _top = parse_vep_or_vcfanno(_p,
                                            is_max=is_max,
                                            _val=_top,
                                            absolute=absolute,
                                            average=average)

    return round(_top, 4)


def process_condel_carol(preds: list):
    """
    Process Condel and CAROL scores
    so that a numeric prediction
    is returned.

    This method is designed to process
    protein predictions made by some
    plugins in VEP, where an outcome
    category is added to the score.
    (e.g. SIFT scores are also returned
    like this by default, although VETA
    expects the dbNSFP format for that tool)

    :param list preds: List of predictions
    :return float: Final prediction
    """

    if len(preds) > 1:
        raise ValueError('Multiple VCF fields were used in the '
                         'config file in a tool that expects a '
                         'single field with predictions made by '
                         'VEP (e.g. Condel, CAROL, SIFT)'
                         'Error: {}'.format(preds))

    _p = preds[0]

    if not _p:
        return np.nan

    # Function expects prediction to start
    # with the 'Deleterious' or 'Neutral' strings
    elif _p.lower().startswith('deleterious') or _p.lower().startswith('neutral'):
        score = float(_p.split("(")[1].split(")")[0])

    else:
        try:
            score = float(_p)
        except ValueError:
            raise ValueError('Unknown string at the beginning '
                             'when processing Condel-like scores: {}'.format(_p))

    return round(score, 4)


def format_spliceai_fields(record, symbol: str):
    """
    Formats spliceAI scores when multiple INFO
    fields are added.

    :param cyvcf2.Record record: Single cyvcf2 record
    :param str symbol: Gene name extracted from SYMBOL
        field added by VEP
    :return str: SpliceAI output formatted in a single
        field.
    """
    fields = [0] * 10
    gname = record.INFO.get("gene_name")
    if gname and gname == symbol or gname and ',' not in gname:
        fields = [record.ALT[0], gname, record.INFO.get("spliceAI_strand"), record.INFO.get("DS_AG"),
                  record.INFO.get("DS_AL"), record.INFO.get("DS_DG"),
                  record.INFO.get("DS_DL"), record.INFO.get("DP_AG"),
                  record.INFO.get("DP_AL"), record.INFO.get("DP_DG"),
                  record.INFO.get("DP_DL")]

    elif gname and ',' in gname:  # if multiple predictions on different genes
        idx = ""
        for g in gname.split(","):
            if g == symbol:
                idx = gname.split(",").index(g)

        if isinstance(idx, str):
            gene_from_id = record.ID.split(":")[0]
            try:
                idx = gname.split(",").index(gene_from_id)

            except ValueError:
                logging.info("Variant ({} {} {}/{}) has multiple SpliceAI scores, and it was not possible "
                             "to extract the proper gene, both from the SYMBOL or ID fields. First ocurrence"
                             " will be employed. SYMBOL:{}, ID:{}, SpliceAI_genes:{}".format(
                    record.CHROM, record.POS, record.REF, record.ALT[0], symbol, record.ID, gname))
                idx = 0

        fields = [record.ALT[0], gname.split(",")[idx], record.INFO.get("DS_AG")[idx], record.INFO.get("DS_AL")[idx],
                  record.INFO.get("DS_DG")[idx], record.INFO.get("DS_DL")[idx],
                  record.INFO.get("DP_AG")[idx], record.INFO.get("DP_AL")[idx],
                  record.INFO.get("DP_DG")[idx], record.INFO.get("DP_DL")[idx]]

    return '|'.join(map(str, [x for x in fields]))


def process_spliceai(preds: pd.Series, check_gene_name: bool = True):
    """
    Processes a list of SpliceAI/CI-SpliceAI
    predictions (SNV or Indel)
    and returns the top scored field.

    It also sets SpliceAI predictions to null
    in cases where the gene of prediction do not 
    match the gene for the selected  VEP consequence.

    :param pd.Series preds: Series with the
    SpliceAI predictions for the given
    SNV or Indel (One of them will always
    be None) and the location of the variant

    :param bool check_gene_name: Retrieve SpliceAI prediction
    on the same gene assigned in the VEP SYMBOL field. Does not
    work for CI-SpliceAI, that outputs gene IDs.
    
    :return float: Top prediction
    """
    tool = "SpliceAI" if check_gene_name else "CI-SpliceAI"
    if all(v is None or v == "." for v in preds[tool]):
        return np.nan
  
    # Iterate over SNVs and Indels:
    _max = 0
    for _p in preds[tool]:

        if _p is not None and _p != ".":
            
            for ind_pred in _p.split(","):
                
                if check_gene_name:
                    if ind_pred.split("|")[1] == preds.SYMBOL:
                        return max(map(float, ind_pred.split("|")[2:6]))
                else:
                    _new_max = max(map(float, ind_pred.split("|")[2:6]))
                    if _new_max > _max:
                        _max = _new_max
                        
            return _max
    return np.nan

def process_conspliceml_like(preds: pd.Series):
    """
    Process ConSpliceML-like scores (any tool with 'something|pred' format)
    to return a single numeric value for variant
    """
    assert len(preds) == 1, "Multiple VCF fields were wrongly provided for a tool that uses the 'conspliceml_like' function."
    
    if all(v is None or v == "." for v in preds):
        return np.nan
    
    if all("|" not in v for v in preds):
        return np.nan
    
    _preds = preds[0].split(",")
    _max = 0
    for _p in _preds:
       
        new_max = round(float(_p.split("|")[1]), 3)
        if abs(new_max) > _max:
            _max = abs(new_max)

    return _max

def process_conspliceml(preds: pd.Series):
    """
    Process ConSpliceML  scores to return
    a single numeric value for the gene
    where variant occurs
    
    :return float: Prediction value
    """
  
    assert len(preds.ConSpliceML) == 1, "Multiple VCF fields were provided in the config for ConSpliceML."
    
    if all(v is None or v == "." for v in preds.ConSpliceML):
        return np.nan

    _preds = preds.ConSpliceML[0].split(",")
    for _p in _preds:
        
        if _p.split("|")[0] == preds.SYMBOL:
            return round(float(_p.split("|")[1]), 3)

    return np.nan

def process_pangolin(preds: list):
    """
    Process Pangolin scores so that the 
    maximum change in splice site usage 
    is returned (as a positive number)
    
    :param list preds: List with the
    Pangolin predictions for the given variant
    
    :return float: Prediction value to be evaluated
    """
    assert len(preds) == 1, "Multiple VCF fields were provided in the config for Pangolin."
    
    if all(v is None or v == "." for v in preds):
        return np.nan
    
    _preds = preds[0].split(",")
    _max = 0
    for _p in _preds:
        s_increase = abs(float(_p.split("|")[1].split(":")[1]))
        s_decrease = abs(float(_p.split("|")[2].split(":")[1]))
        new_max = abs(max(s_increase, s_decrease))
        if new_max > _max:
            _max = new_max

    return _max

def process_spip(preds: list):
    """
    Process SPiP scores so that the 
    the risk of the variant impacting
    splicing is returned
    
    :param list preds: List with the
    SPiP predictions for the given variant
    
    :return float: Prediction value to be evaluated
    """
    assert len(preds) == 1, "Multiple VCF fields were provided in the config for SPiP."
        
    if all(v is None or v == "." for v in preds):
        return np.nan
    
    _preds = preds[0].split(",")
    _max = 0
    for _p in _preds:
        # T|NM_001142771:g.55626401:G>A|Alter by SPiCE|98.41 % [91.47 % - 99.96 %]|1.000|-|55626401|substitution|G>A|Intron 28| ...
        #ci = float(_p.split("|")[3].split()[0])
        score = float(_p.split("|")[4])
        if score > _max:
            _max = score
    return _max
    
def process_scap(preds: pd.Series):
    """
    Process S-CAP scores so that
    different reference thresholds
    are taken into account
    for pathogenicity prediction

    If prediction is above the reference
    threshold, 1 is summed up to the final
    score. To evaluate predictions, S-CAP scores
    are then compared to 1. Note that ROC
    analysis won't be performed on such
    scores due to this transformation.

    :param pd.Series preds: Series with the
    S-CAP predictions for the given variant
    as well as its location (which won't be used)

    :return float: Prediction value
    """

    is_not_core_region = ["3intronic", "exonic", "5extended", "5intronic"]
    thresholds = {"3intronic": 0.006,
                  "exonic": 0.009,
                  "5extended": 0.005,
                  "5intronic": 0.006,
                  "5core_dominant": 0.034,
                  "5core_recessive": 0.367,
                  "3core_dominant": 0.033,
                  "3core_recessive": 0.264
                  }

    if len(preds) > 1:
        raise ValueError('Multiple VCF fields were used in the '
                         'config file when setting up S-CAP. However '
                         'this tool expects a single field with several '
                         'values to be parsed according the region defined. Error: {}'.format(preds))

    _p = preds[0]

    if not _p or _p == ".":
        return np.nan

    elif "COMMON" in _p:
        return 0

    else:
        try:
            fields = _p.split(":")
            if fields[1] in is_not_core_region:
                # rawscore
                if float(fields[2]) >= thresholds[fields[1]]:
                    return float(fields[3]) + 1
                else:
                    return float(fields[3])

            elif fields[1] == "5core":
                # We do not know whether a variant is
                # dominant and recessive, thus we test
                # for both dom and rec thresholds.
                # If any greater than threshold, prediction is
                # pathogenic
                if float(fields[4]) >= thresholds['5core_dominant']:
                    return float(fields[4]) + 1

                elif float(fields[6]) >= thresholds['5core_recessive']:
                    return float(fields[6]) + 1

                else:
                    return float(fields[4])

            elif fields[1] == "3core":
                if float(fields[4]) >= thresholds['3core_dominant']:
                    return float(fields[4]) + 1

                elif float(fields[6]) >= thresholds['3core_recessive']:
                    return float(fields[6]) + 1

                else:
                    return float(fields[4])

        # if prediction is like '1:exonic:.:.:.:.:.:.'
        except ValueError:
            return np.nan


def process_trap(preds: pd.Series):
    """
    Process TrAP scores so that
    different coding and intronic
    thresholds are taken into account
    for pathogenicity prediction

    If prediction is above the reference
    threshold, 1 is summed up to the final
    score. To evaluate predictions, TraP scores
    are then compared to 1. Note that ROC
    analysis won't be performed on such
    scores due to this transformation.

    :param pd.Series preds: Series with the
    TraP predictions for the given variant
    as well as its location

    :return float: Prediction value
    """
   
    if all(v is None for v in preds.TraP):
        return np.nan

    if len(preds.TraP) > 1:
        raise ValueError('Multiple VCF fields were used in the '
                         'config file when setting up TraP. However '
                         'this tool expects a single field with numeric '
                         'predictions. Error: {}'.format(preds.TraP))

    _p = preds.TraP[0]
    score = ""
    try:
        score = float(_p)
    except ValueError:
        if "," in _p:
            score = max([float(v) for v in _p.split(',') if _p.strip() != '.'])

    if preds.location in ["splice_site", "splice_region", "intronic",
                          "deep_intronic", "5primeUTR", "3primeUTR"]:
        return score + 1 if score > 0.289 else score

    else:
        return score + 1 if score > 0.416 else score


def process_kipoi_tools(x: list, _max: bool = True):
    """
    Process scores given by Kipoi, which usually
    comprise several predictions, one for each
    genomic feature present in the gft used
    that overlaps with the input variant

    For each element (if multiple VCF fields
    define the score) in x, one single prediction
    is returned. Max value from each element is
    retrieved by default (_max).

    :param x: List of lists with the scores
    for a single variant
    :param bool _max: Whether max value should
    be retrieved. Default: `True`. If `False`,
    min is used.

    :return Union[float, List]: Top predictions
    for each element. If tool is defined by one
    single VCF field, returns a single value
    """

    def _get_prediction(elem: Union[str, float]):
        """
        Returns top prediction for a single
        element

        :param str elem: Score(s)
        :return float:
        """
        if isinstance(elem, float):
            return abs(elem) if _max is True else elem

        if not elem or elem == "." or elem == "None":
            return np.nan

        elif "," in elem and _max is True:
            return round(max([abs(float(v)) for v in elem.split(",")]), 4)

        elif "," in elem:
            return round(min([abs(float(v)) for v in elem.split(",")]), 4)

        elif _max is True:
            return abs(float(elem))

        else:
            return float(elem)

    out = []
    for e in x:
        out.append(_get_prediction(e))

    return out[0] if len(out) == 1 else out

def process_labranchor(preds: list):
    """
    Process LaBranchoR scores so that the 
    the risk of the variant impacting
    splicing is returned
    
    :param list preds: List with the
    LaBranchoR predictions for the given variant
    
    :return float: Prediction value to be evaluated
    """
    assert len(preds) == 1, "Multiple VCF fields were provided in the config for LaBranchoR."
    
    if all(v is None or v == "." for v in preds):
        return np.nan
    
    _preds = preds[0].split(",")
    _max = 0
    for _p in _preds:
        positions = map(float, _p.split("|"))
        score = abs(max(positions, key=abs))
        if score > _max:
            _max = score
    return _max


def process_absplice(preds: list):
    """
    Process AbSplice_DNA scores so that the 
    the tissue with the highest score is returned


    :param list preds: List with the per-tissue
    AbSplice_Dna predictions for the given variant
    
    :return float: Prediction value to be evaluated
    """
    assert len(preds) == 1, "Multiple VCF fields were provided in the config for AbSplice_DNA."
    
    if preds[0] is None:
        return np.nan
    
    # If variant overlaps multiple genes, there are multiple predictions
    _preds = preds[0].split(",")
    
    _max = -1
    for _p in _preds:
        
        # index 0 is the gene name
        _p = _p.split("|")[1:]
        
        top_score = max([-1 if x == '' else float(x) for x in _p])
        if top_score > _max:
            _max = top_score
    
    if _max == -1:
        return np.nan 
    return _max


import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns 


def set_style():
    
    plt.style.use(['seaborn-white', 'seaborn-paper'])
    matplotlib.rc("font", family="Times New Roman")
    
    # This sets reasonable defaults for font size for
    # a figure that will go in a paper
    sns.set_context("paper")
    
    # Set the font to be serif, rather than sans
    sns.set(font='serif')
    
    # Make the background white, and specify the
    # specific font family
    sns.set_style("white", {
        "font.family": "serif",
        "font.serif": ["Times", "Palatino", "serif"]
    })


def set_size(fig, s=10):
    fig.set_size_inches(8, s/3+1)
    plt.tight_layout()


def format_fn(tick_val, tick_pos):
    return abs(tick_val)


def nonna(col):
    return col.isnull().sum() == col.shape[0]



from typing import List
import numpy as np
import pandas as pd


def apply_tool_predictions(df: pd.DataFrame, thresholds: List):
    """
    Evaluate tools predictions according to the
    reference thresholds

    :param pd.DataFrame df: Input df
    :param List thresholds: List with the
    reference thresholds

    :return pd.DataFrame: Df with predictions
    for each tool
    """
    for tool, direction, threshold, *args in thresholds:
        if tool not in df.columns:
            # logging.info(tool + " scores were not processed. This tool may not be"
            #                     " in the original VCF header or did not score any"
            #                     " variant.")
            continue

        if direction == ">":
            classification = lambda x: pd.isna(x) and np.nan or x >= threshold
        else:
            classification = lambda x: pd.isna(x) and np.nan or x <= threshold

        df[tool + '_prediction'] = df[tool].apply(classification)
    return df



import logging
from typing import Union
import numpy as np
from gplearn.genetic import SymbolicClassifier
from sklearn import tree
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis, QuadraticDiscriminantAnalysis
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.exceptions import NotFittedError, ConvergenceWarning
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV, cross_validate
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import Normalizer, StandardScaler
from sklearn.svm import LinearSVC, SVC
from sklearn.utils._testing import ignore_warnings


class Classifiers(object):
    """
    Class to test multiple classifiers using different hyperparameters
    """

    def __init__(self, x: np.array, y: np.array, scoring: dict, refit_metric: str):
        """
        Initialization with the scikit learn scoring metrics desired

        :param np.array x: Training data
        :param np.array y: Labels

        :param dict scoring: Dict with predefined metrics. Keys must
            be metric names, while values should be predefined metric
            string recognized by sklearn

        :param srt refit_metric: Refers to the metric to use for model
            selection after best parameter selection. Then, the model
            is refitted using all the data and the selected model. Value
            must be present in `scoring` keys.
        """
        self.x = x
        self.y = y.reshape(-1)
        self.scoring = scoring
        assert refit_metric in self.scoring.keys(), "Refit metric must exist in scoring keys."
        self.refit_metric = refit_metric

    def knn(self) -> Pipeline:
        """
        Creates a pipeline for a K-nearest neighbors classifier

        Since it uses distances to learn, a Normalization preprocessing
        step is also included.

        :return Pipeline: returns a Pipeline for the best estimator
        """

        pipeline = Pipeline(steps=[('normalization', Normalizer()),
                                   ('knn', KNeighborsClassifier())])
        params_grid = {'normalization__norm': ['l1', 'l2'],
                       'knn__n_neighbors': [3, 5, 10]}
        return self.do_grid_search("knn", pipeline, params_grid)

    def naive_bayes(self) -> Pipeline:
        """
        Creates a pipeline for a Gaussian Naive Bayes classifier

        Since it uses distances to learn, a Normalization preprocessing
        step is also included.

        :return Pipeline: returns a Pipeline for the fitted estimator
        """
        pipeline = Pipeline(steps=[('NaiveBayes', GaussianNB())])
        return self.do_grid_search("NaiveBayes", pipeline)

    @ignore_warnings(category=NotFittedError)
    def linear_discrimant_analysis(self) -> Pipeline:
        """
        Creates a pipeline for a Linear Discriminant analysis

        :return Pipeline: returns a Pipeline for the best estimator
        """

        pipeline = Pipeline(steps=[('lda', LinearDiscriminantAnalysis())])
        params_grid = {'lda__solver': ['svd', 'lsqr']}
        return self.do_grid_search("lda", pipeline, params_grid, njobs=1)

    @ignore_warnings(category=UserWarning)
    def quadratic_discriminant_analysis(self) -> Pipeline:
        """
        Creates a pipeline for a Quadratic Discriminant analysis

        :return Pipeline: returns a Pipeline for the best estimator
        """
        pipeline = Pipeline(steps=[('qda', QuadraticDiscriminantAnalysis())])
        params_grid = {'qda__reg_param': [0.0, 0.2, 0.4, 0.6, 0.8, 1],
                       'qda__store_covariance': [True]}

        return self.do_grid_search("qda", pipeline, params_grid, njobs=1)

    def decision_tree(self) -> Pipeline:
        """
        Creates a pipeline for a Decision Tree

        :return Pipeline: returns a Pipeline for the best estimator
        """
        pipeline = Pipeline(steps=[('dt', tree.DecisionTreeClassifier(random_state=0))])

        params_grid = {'dt__criterion': ['gini', 'entropy'],
                       'dt__splitter': ['best', 'random'],
                       'dt__max_depth': [3, 5, 7],
                       'dt__min_samples_split': [2, 4, 6],
                       'dt__min_samples_leaf': [1, 2, 3, 4],
                       'dt__max_features': [None, "auto", "sqrt"]}

        return self.do_grid_search("dt", pipeline, params_grid)

    def random_forest(self) -> Pipeline:
        """
        Creates a pipeline for a Random Forest

        :return Pipeline: returns a Pipeline for the best estimator
        """
        pipeline = Pipeline(steps=[('rf', RandomForestClassifier(random_state=0))])

        params_grid = {'rf__n_estimators': [10, 50, 100],
                       'rf__criterion': ['gini', 'entropy'],
                       'rf__max_depth': [5, 7],
                       'rf__min_samples_split': [2, 4],
                       'rf__min_samples_leaf': [2, 4],
                       'rf__max_features': ["auto"],
                       'rf__bootstrap': [True],
                       # 'rf__ccp_alpha': [0.0, 0.1],
                       'rf__max_samples': [None, 0.8, 0.9]}

        return self.do_grid_search("rf", pipeline, params_grid)

    def ada_boost(self) -> Pipeline:
        """
        Creates a pipeline for a AdaBoost

        :return Pipeline: returns a Pipeline for the best estimator
        """
        pipeline = Pipeline(steps=[('ab', AdaBoostClassifier(random_state=0))])

        params_grid = {'ab__n_estimators': [10, 50, 100],
                       'ab__learning_rate': [0.5, 1]
                       }

        return self.do_grid_search("svm", pipeline, params_grid)

    @ignore_warnings(category=ConvergenceWarning)
    def svm(self) -> Pipeline:
        """
        Creates a pipeline for Support Vector Machines

        :return Pipeline: returns a Pipeline for the best estimator
        """
        pipeline = Pipeline(steps=[('scaler', StandardScaler()),
                                   ('svm', LinearSVC(random_state=0))])

        # Pipeline that tests multiple SVM classifiers
        params_grid = [{'svm': (SVC(random_state=0),),
                        'svm__C': (0.001, 0.01, 0.1, 1, 10),
                        'svm__kernel': ('poly', 'rbf', 'sigmoid')},

                       {'svm': (LinearSVC(random_state=0),),
                        # 'svm__penalty': ('l1', 'l2'),
                        'svm__loss': ('hinge', 'squared_hinge'),
                        'svm__C': ([0.001, 0.01, 0.1, 1, 10]),
                        'svm__max_iter': (1500, 2000),
                        }
                       ]

        return self.do_grid_search("svm", pipeline, params_grid)

    def logistic_regression(self) -> Pipeline:
        """
        Creates a pipeline for Support Vector Machines

        :return Pipeline: returns a Pipeline for the best estimator
        """
        pipeline = Pipeline(steps=[('scaler', StandardScaler()),
                                   ('lr', LogisticRegression(random_state=0))])

        params_grid = {'lr__penalty': ['l1', 'l2'],
                       'lr__solver': ['liblinear']}

        return self.do_grid_search("lr", pipeline, params_grid)

    def gp(self) -> Pipeline:
        """
        Creates a pipeline for Genetic programming

        :return Pipeline: returns a Pipeline for the best estimator
        """
        pipeline = Pipeline(steps=[('scaler', StandardScaler()),
                                   ('gp', SymbolicClassifier(random_state=0))])

        params_grid = {'gp__generations': [10, 50, 100]}

        return self.do_grid_search("gp", pipeline, params_grid)

    def do_grid_search(self, name: str,
                       pipeline: Pipeline,
                       params_grid: Union[dict, list] = None,
                       njobs: int = -1) -> Pipeline:
        """
        Perform grid search for a given pipeline

        :param str name: Name of the estimator
        :param pipeline: Pipeline for a given estimator
        :param dict params_grid: Parameters to test in the
            grid search for the provided pipeline.
        :param int njobs: Number of jobs to run in parallel.
            Default: `-1`, use all available cpus.

        :return Pipeline: Returns a Pipeline for the best estimator
            so that it can be evaluated on test data (making predictions)
        """
        # If there are no parameters to tune, apply CV directly
        if params_grid is None:
            res = cross_validate(pipeline, self.x, self.y,
                                 cv=10,
                                 scoring=self.scoring,
                                 n_jobs=njobs)

            cv_score = np.mean(res['test_' + self.refit_metric])
            logging.info("Mean cross validated score ({}) for the {} estimator: {:.3f}".format(self.refit_metric,
                                                                                               name,
                                                                                               cv_score))
            # Fit all training data
            return pipeline.fit(self.x, self.y), None, cv_score

        else:
            # StratifiedKFold is performed for each of parameter combination
            # with ignore_warnings(category=[ConvergenceWarning, FitFailedWarning]):
            search = GridSearchCV(pipeline,
                                  params_grid,
                                  scoring=self.scoring,
                                  refit=self.refit_metric,
                                  cv=10,
                                  n_jobs=njobs,
                                  )

            search.fit(self.x, self.y)
            logging.info("Best parameters found for {} estimator: {}".format(name, search.best_params_))
            logging.info("Mean cross validated score ({}) of the best {} estimator: {:.3f}".format(self.refit_metric,
                                                                                                   name,
                                                                                                   search.best_score_))
            return search.best_estimator_, search.best_params_, search.best_score_

    def train(self, name: str) -> Pipeline:
        """
        Run pipeline for a given algorithm

        :param name:
        :return:
        """
        if name == "KNN":
            return self.knn()

        elif name == "Naive_Bayes":
            return self.naive_bayes()

        elif name == "LDA":
            return self.linear_discrimant_analysis()

        elif name == "QDA":
            return self.quadratic_discriminant_analysis()

        elif name == "Decision_Tree":
            return self.decision_tree()

        elif name == "Random_Forest":
            return self.random_forest()

        elif name == "Ada_Boost":
            return self.ada_boost()

        elif name == "SVM":
            return self.svm()

        elif name == "Logistic_Regression":
            return self.logistic_regression()

        elif name == "Genetic_Programming":
            return self.gp()



import logging
import numpy as np
import pandas as pd
from typing import Union, List
from collections import defaultdict
from preprocessing.utils_tools import ratio
from sklearn.metrics import auc, average_precision_score, roc_auc_score, roc_curve, precision_recall_curve, matthews_corrcoef


def generate_statistics(df: pd.DataFrame,
                        statistics: defaultdict,
                        filter_location: str,
                        tool: str,  
                        is_single_label: bool = False,
                        f_beta: Union[List, int] = None):
    """
    Compute all stats for a single tool

    :param pd.DataFrame df: Input df with predictions
        compared to the reference threshold
    :param defaultdict statistics: Dict with outputs stats
        to be updated
    :param str filter_location: Location filter
    :param str tool: Tool to compute stats
    :param bool is_single_label: Whether df refers
        to an analysis in the inspect mode with a
        single label. Default: `False`.
    :param Union[List, int] f_beta: Calculate F beta score for the given
    beta value(s). Default: `None`
    :return defaultdict: Updated dict
    """
    # s_df = df[~df[tool + '_prediction'].isnull()]
    total = df.shape[0]
    s_df = df[~df[tool].isnull()]
    statistics['filter'].append(filter_location)
    statistics['tool'].append(tool)

    if is_single_label is False:
        tp = np.sum(s_df['label'].eq(s_df[tool + '_prediction']) & s_df['label'])
        tn = np.sum(s_df['label'].eq(s_df[tool + '_prediction']) & ~s_df['label'])

        fp = np.sum(s_df['label'].ne(s_df[tool + '_prediction']) & ~s_df['label'])
        fn = np.sum(s_df['label'].ne(s_df[tool + '_prediction']) & s_df['label'])

        mp = np.sum(df[tool + '_prediction'].isnull() & df['label'])
        mn = np.sum(df[tool + '_prediction'].isnull() & ~df['label'])
        nan = np.sum(df[tool + '_prediction'].isnull())
        correct = tp + tn

        precision = ratio(tp, tp + fp)
        recall = ratio(tp, tp + fn)
        coverage = ratio(tp + tn + fp + fn, total)
        accuracy = ratio(correct, (total - nan))
        
        #mcc = ((tp * tn) - (fp * fn)) / (sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn)))
        mcc = round(matthews_corrcoef(s_df.label.astype(int), s_df[tool + '_prediction'].astype(int)), 3)
        normalized_mcc = (mcc + 1) / 2
        weighted_norm_mcc = round(normalized_mcc * coverage, 3)
        
        statistics['precision'].append(precision)
        statistics['specificity'].append(ratio(tn, tn + fp))
        statistics['sensitivity'].append(recall)
        statistics['mcc'].append(mcc)
        statistics['norm_mcc'].append(normalized_mcc)
        statistics['weighted_norm_mcc'].append(weighted_norm_mcc)
        statistics['tp'].append(tp)
        statistics['fp'].append(fp)
        statistics['tn'].append(tn)
        statistics['fn'].append(fn)
        statistics['mp'].append(mp)
        statistics['mn'].append(mn)

        statistics['scored_p'].append(np.sum(s_df['label']))
        statistics['scored_n'].append(np.sum(~s_df['label']))
        statistics['total_p'].append(np.sum(df['label']))
        statistics['total_n'].append(np.sum(~df['label']))

        try:
            statistics['F1'].append(round(2 * (precision * recall) /
                                          (precision + recall), 3))

            statistics['weighted_F1'].append(round((2 * (precision * recall) /
                                                    (precision + recall)) * coverage, 3))

        except ZeroDivisionError:
            statistics['F1'].append(0)
            statistics['weighted_F1'].append(0)

        if f_beta is not None:
            if isinstance(f_beta, int):
                f_beta = [f_beta]

            for _b in f_beta:
                try:
                    statistics['Fbeta_' + str(_b)].append(round((1.0 + _b ** 2) * (precision * recall) /
                                                                ((precision * _b ** 2) + recall), 3))

                    statistics['weighted_Fbeta_' + str(_b)].append(
                        round(coverage * (1.0 + _b ** 2) * (precision * recall) /
                              ((precision * _b ** 2) + recall), 3))
                except ZeroDivisionError:
                    statistics['Fbeta_' + str(_b)].append(0)
                    statistics['weighted_Fbeta_' + str(_b)].append(0)

    else:
        correct = np.sum(s_df['label'].eq(s_df[tool]))
        nan = np.sum(df[tool].isnull())
        accuracy = ratio(correct, (total - nan))
        coverage = ratio((total - nan), total)

    weighted_accuracy = round(accuracy * coverage, 3) 
    statistics['total'].append(total)
    statistics['correct'].append(correct)
    statistics['nan'].append(nan)
    statistics['fraction_nan'].append(ratio(nan, total))
    statistics['coverage'].append(coverage)
    statistics['accuracy'].append(accuracy)
    statistics['weighted_accuracy'].append(weighted_accuracy)

    return statistics


def do_roc_analysis(_data: pd.DataFrame,
                    name: str,
                    higher_is_better: bool = True):
    """
    Perform ROC analysis a given tool

    :param pd.Dataframe data: Scores for a given tool
    with corresponding labels
    :param str name: Tool name
    :param bool higher_is_better: If pathogenicity is predicted when prediction is higher than reference threshold
    """
    
    # ROC analysis is not performed for 
    # S-CAP (it uses multiple reference thresholds) and 
    # cVEP (predictions are discrete categories that VETA converts to fixed floats)

    if name.lower() in ['scap', 's-cap', 'cvep']:
        return

    data = _data.copy()
    max_thr = data[name].max()  # + (df_tool[tool].max()) - df_tool[tool].min()) * 0.001
    min_thr = data[name].min()  # - (df_tool[tool].max()) - df_tool[tool].min()) * 0.001

    if pd.isnull(max_thr):
        logging.info("{} has no predicted variants".format(name))
        return

    elif max_thr == min_thr:
        logging.info("{} has the the same min/max score".format(name))
        return

    if higher_is_better is False:
        # Reverse scores, so that higher is better
        data['y_pred'] = data[name] * (-1)

    else:
        data['y_pred'] = data[name]
    
    data = data[['label', 'y_pred']].sort_values('y_pred')
    data['rank'] = np.arange(len(data))
    data['rank_score'] = data['rank'] / len(data)

    fpr, tpr, thresh_roc = roc_curve(data.label, data.rank_score)
    precision, recall, thresh_pr = precision_recall_curve(data.label, data.rank_score)

    label = name + "(n=" + str(data.shape[0]) + ","

    if len(thresh_roc) > 400:
        idx = np.round(np.linspace(0, len(thresh_roc) - 1, 200)).astype(int)
        thresh_roc = thresh_roc[idx]
        tpr = tpr[idx]
        fpr = fpr[idx]

    precision = precision[:-1]
    recall = recall[:-1]
    if len(thresh_pr) > 400:
        idx = np.round(np.linspace(0, len(thresh_pr) - 1, 200)).astype(int)
        thresh_pr = thresh_pr[idx]
        precision = precision[idx]
        recall = recall[idx]

    roc_curve_data = [label, list(thresh_roc), list(tpr), list(fpr)]
    pr_curve_data = [label, list(thresh_pr), list(recall), list(precision)]

    return roc_curve_data, pr_curve_data, auc(fpr, tpr), auc(recall, precision)



import logging
import os
import numpy as np
import pandas as pd
import multiprocessing
from functools import partial
from collections import OrderedDict, defaultdict
from tqdm import tqdm
from sklearn.utils import resample
from plots.plots_threshold_analysis import plot_optimal_thresholds, plot_bootstrap_thresholds
from preprocessing.osutils import ensure_folder_exists
from preprocessing.utils_tools import ratio


def fbeta_at_thr(df, tool, beta_values, stats):
    """
    Calculate stats at a single threshold
    """
    tp = np.sum(df['label'].eq(df[tool + '_prediction']) & df['label'])
    tn = np.sum(df['label'].eq(df[tool + '_prediction']) & ~df['label'])
    fp = np.sum(df['label'].ne(df[tool + '_prediction']) & ~df['label'])
    fn = np.sum(df['label'].ne(df[tool + '_prediction']) & df['label'])
    total = df.shape[0]
    correct = tp + tn
    specificity = ratio(tn, tn + fp)
    precision = ratio(tp, tp + fp)
    recall = ratio(tp, tp + fn)
    accuracy = ratio(correct, total)
        
    for _b in beta_values:
        try:
            stats['Fbeta_' + str(_b)].append(round((1.0 + _b ** 2) * (precision * recall) /
                                                        ((precision * _b ** 2) + recall), 3))
        except ZeroDivisionError:
            stats['Fbeta_' + str(_b)].append(0)

    stats['accuracy'].append(accuracy)
    stats['specificity'].append(specificity)
    stats['sensitivity'].append(recall)
    return stats
                    
def classify_at_threshold(df, thr_range, **kwargs):
    """
    Iterate over each putative threshold, classify, and generate metrics
    """
    stats = defaultdict(list)
    for threshold in thr_range:
        if kwargs['direction'] == ">":
            def classification_f(
                x): return x == np.nan and np.nan or x >= threshold
        else:
            def classification_f(
                x): return x == np.nan and np.nan or x <= threshold

        df[kwargs['tool'] + '_prediction'] = df[kwargs['tool']].map(classification_f)
        
        stats = fbeta_at_thr(df, kwargs['tool'], kwargs['beta_values'], stats)
    
    # If evaluating with best thr after bootstrapping
    if len(thr_range) == 1:
        return stats
    
    new_t = {}
    f_betas = defaultdict(list)
    for _beta in kwargs['beta_values']:
        _res_at_each_t = stats['Fbeta_' + str(_beta)]
        f_betas[_beta] = _res_at_each_t
        new_t[_beta] = thr_range[_res_at_each_t.index(max(_res_at_each_t))]
        kwargs['final_thresh'][kwargs['tool']].append((round(float(new_t[_beta]), 3), max(_res_at_each_t)))
        
    return new_t, stats, f_betas

def _calculate_ci(bootst_thr: pd.DataFrame, **kwargs):
    """
    Calculates best threshold by averaging the
    list of thresholds obtained at each bootstrap iteration.
    Additionally, calculates 95% CI and returns Fbeta at best
    threshold
    
    :param pd.DataFrame bootst_thr: Df with list of bootstrap threshold
    for each beta value
    """
    # Info from threshold analysis of the data sample (before bootstrapping)
    ref_thr_info = kwargs['final_thresh'][kwargs['tool']][0]
    adj_thr_info = kwargs['final_thresh'][kwargs['tool']][1:]

    bootst_thr = bootst_thr.iloc[0]
    out = [ref_thr_info]
    for i, b in enumerate([0.5, 1, 1.5]):
  
        # Deriving best thresh by averaging bootstrap statistic values
        # best_thr = round(np.mean(bootst_thr['thr_f{}'.format(b)]), 3)
        #stats = classify_at_threshold(kwargs['df_per_tool'], [best_thr], **kwargs)   
        #fbeta_at_best_thr = stats['Fbeta_{}'.format(b)][0]
        
        dist_thr = bootst_thr['thr_f{}'.format(str(b))]

        # Empirical bootstrap confidence interval
        #ci95_up = np.mean(dist_thr) + 1.96 * np.std(dist_thr) / np.sqrt(len(dist_thr))
        #ci95_down = np.mean(dist_thr) - 1.96 * np.std(dist_thr) / np.sqrt(len(dist_thr))
        
        # Boostrap percentile method
        ci = pd.Series(dist_thr).quantile([0.025, 0.975])
        ci95_down = round(ci[0.025], 3)
        ci95_up = round(ci[0.975], 3)
        
        # Add percentiles to output
        best_thr = adj_thr_info[i][0]
            
        adj_with_percent = ('{} ({},{})'.format(best_thr, str(ci95_down), str(ci95_up)),
                           '{}'.format(adj_thr_info[i][1]))

        out.append(adj_with_percent)
        plot_bootstrap_thresholds(dist_thr, kwargs['tool'], b, kwargs['_loc'], best_thr, kwargs['ref_thresh'], ci95_down, ci95_up, kwargs['outdir'])

    kwargs['final_thresh'][kwargs['tool']] = out

def _do_bootstapping(_iter: int, **kwargs):

    df_per_tool = kwargs['df_per_tool']
    tool = kwargs['tool']
    n_of_points = kwargs['n_of_points']

    max_thr = df_per_tool[tool].max()
    min_thr = df_per_tool[tool].min()

    if pd.isnull(max_thr) or pd.isnull(min_thr) or max_thr == min_thr:
        logging.info("Something strange in max/min thresholds {} {} "
                     "{}".format(tool, max_thr, min_thr))
        return

    data = resample(df_per_tool, 
                    replace=True, 
                    n_samples=df_per_tool.shape[0], 
                    stratify=df_per_tool.label)
    
    #n_of_points = random.randint(n_of_points, n_of_points + 500)
    #step = (max_thr - min_thr) / float(n_of_points)
    #thr_range = np.arange(min_thr, max_thr, step)

    thr_range = np.sort(np.random.uniform(min_thr, max_thr, size=n_of_points))
    new_t, _, _ = classify_at_threshold(data, thr_range, **kwargs)
    return new_t


def perform_threshold_analysis(dataset: pd.DataFrame,
                               location_filters: list,
                               threshold_list: list,
                               tools_config: dict,
                               outdir: str,
                               f1_at_ref_thresh: dict,
                               n_of_points: int = 100,
                               beta_values: list = [0.5, 1, 1.5],
                               do_bootstrapping: bool = False):
    """
    Do threshold analysis using Clinvar as ground
    truth dataset for threshold evaluation.

    Additionally, several precision/recall trade-offs
    are tested to find the most appropriate threshold
    per tool that maximimizes the number of true positives
    detected.

    :param pd.DataFrame dataset: Input dataset
    :param list location_filters: List of location filters
    :param list threshold_list: List of thresholds
    :param dict tools_config: Dict with config for each tool
    :param str outdir: Output directory
    :param dict f1_at_ref_thresh: Dict with F1 scores
    using the reference thresholds
    :param n_of_points: Default: `100`
    :param list beta_values: List of values to use as beta
        in the F_beta score calculation (precision/recall
        tradeoff). Default: `[0.5, 1, 1.5]`
    :param bool do_bootrstrapping: Set this to True to 
    generate confidence intervals
    :return dict: New thresholds for all the tools at all locations
    """
    logging.info("---------------------------")
    logging.info("Starting threshold analysis")
    logging.info("---------------------------")
    outdir = os.path.join(outdir, "thresholds_analysis")
    ensure_folder_exists(outdir)

    available_for_threshold_analysis = [
        'all', 'coding', 'missense', 'splice_site', 'splice_region', 'intronic', 'deep_intronic']

    # Output dict initialization
    thresholds_to_return = {}
    for beta in beta_values:
        thresholds_to_return[beta] = {}
        for _loc in location_filters:
            if _loc not in available_for_threshold_analysis:
                continue
            thresholds_to_return[beta][_loc] = {}

    for _loc in location_filters:

        if _loc not in available_for_threshold_analysis:
            continue
        
        elif _loc == 'all':
            df_f = dataset.copy()
        
        else:
            df_f = dataset[dataset.location == _loc].copy()
            
        print()
        logging.info("-------------------------")
        logging.info("Looking at {} variants (N={}, {} pos and {} neg)".format(_loc,
                                                                               df_f.shape[0],
                                                                               df_f[df_f.label].shape[0],
                                                                               df_f[~df_f.label].shape[0]))
        logging.info("-------------------------")
        # Do threshold analysis only if dataset has a minimum size
        if df_f.shape[0] < 100:
            logging.info('No minimum number of variants required (N=100)')
            continue

        if not all(i >= 50 for i in df_f.label.value_counts().tolist()):
            logging.warning(
                'No minimum number of variants in the minority class required (N=50)')

            continue

        final_thresholds = OrderedDict()
        for tool, direction, reference_threshold, *args in threshold_list:

            # Threshold analysis is not performed for
            # S-CAP (it uses multiple reference thresholds) and

            # cVEP (predictions are discrete categories that VETA converts to fixed floats)
            if tool not in df_f.columns or tool.lower() in ['scap', 's-cap', 'cvep']:
                continue

            # Do threshold analysis for tools with sufficient prediction power
            df_per_tool = df_f.loc[pd.notnull(df_f[tool]), ][[
                tool, 'label']].copy()

            # Transform TraP scores for proper threshold analysis
            if tool == "TraP":
                df_per_tool['TraP'] = df_per_tool['TraP'].apply(
                    lambda x: x - 1 if x > 1 else x)

                # If all variants, it is not possible to distinguish coding/non-coding thresholds
                if _loc in ["coding", "missense"]:
                    reference_threshold = 0.416
                else:
                    reference_threshold = 0.289

            ratio_predicted = df_per_tool.shape[0] / df_f.shape[0]

            if df_per_tool.shape[0] < 100:
                logging.info(
                    '{} did not predict the minimum number of variants required (N=100)'.format(tool))
                continue
            
            if not all(i >= 50 for i in df_per_tool.label.value_counts().tolist()):
                logging.warning('{} did not predict the minimum number of variants in the minority class required (N=50)'.format(tool))
                continue
            
            if ratio_predicted < 0.5:
                logging.info("{} predicted less than 50% of variants (N={} predictions). "
                             "Threshold analysis will be performed, but be cautious in the "
                             "interpretation.".format(tool, df_per_tool.shape[0]))

            kwargs = {'df_per_tool': df_per_tool,
                      'tool': tool,
                      'n_of_points': n_of_points,
                      'direction': direction,
                      'f1_at_ref_thresh': f1_at_ref_thresh,
                      '_loc': _loc,
                      'beta_values': beta_values,
                      'ref_thresh': reference_threshold,
                      'final_thresh': final_thresholds,
                      'outdir': outdir}
            kwargs['final_thresh'][tool] = [(kwargs['ref_thresh'], kwargs['f1_at_ref_thresh'][_loc][tool])]
            max_thr = df_per_tool[tool].max()
            min_thr = df_per_tool[tool].min()
 
            if pd.isnull(max_thr) or pd.isnull(min_thr) or max_thr == min_thr:
                logging.info("Something strange in max/min thresholds {} {} "
                    "{}".format(tool, max_thr, min_thr))


            step = (max_thr - min_thr) / float(n_of_points)
            thr_range = np.arange(min_thr, max_thr, step)

            best_t, stats, fbetas = classify_at_threshold(df_per_tool, thr_range, **kwargs)

            for v in [False, True]:
                plot_optimal_thresholds(tool,
                                        _loc,
                                        thr_range,
                                        stats['accuracy'],
                                        stats['sensitivity'],
                                        stats['specificity'],
                                        fbetas,
                                        reference_threshold,
                                        best_t,
                                        outdir,
                                        simple=v)
            
            if do_bootstrapping:
                logging.info("Bootstrapping analysis for {}".format(tool))
                thr_f05, thr_f1, thr_f15 = [], [], []
                with multiprocessing.Pool() as p:
                    new_t = list(tqdm(p.imap(partial(_do_bootstapping, **kwargs), range(1000)), total=1000))    
                    for t in new_t: 
                        thr_f05.append(t[0.5])
                        thr_f1.append(t[1])
                        thr_f15.append(t[1.5])
                    
                cols = ['tool', 'thr_f0.5', 'thr_f1', 'thr_f1.5']
                thr_df = pd.DataFrame([[tool, thr_f05, thr_f1, thr_f15]], columns=cols)
                _calculate_ci(thr_df, **kwargs)

        _outdir_new_thresh = os.path.join(outdir, "new_thresholds")
        _outdir_new_config = os.path.join(outdir, "new_configs")
        ensure_folder_exists(_outdir_new_config)
        ensure_folder_exists(_outdir_new_thresh)


        ###########################
        ### Write new thresholds ##
        ###########################
        final_thresholds = kwargs['final_thresh']
        with open(os.path.join(_outdir_new_thresh, "proposed_thresholds_{}.tsv".format(_loc)), 'w') as out:
            out.write("Tool\tReference_threshold\tF1_score_at_ref_thresh\t{}\n".format("\t".join("Thresh_beta_{}".format(b) +
                                                                                                 "\t" +
                                                                                                 "Fbeta_{}".format(b)
                                                                                                 for b in beta_values)))

            for tool, new_thresholds in final_thresholds.items():
                out.write(
                    "{}\t{}\n".format(tool, '\t'.join(["{}\t{}".format(str(t[0]), str(t[1])) for t in new_thresholds])))

        for i, beta in enumerate(beta_values):

            config = open(os.path.join(_outdir_new_config,
                          "new_config_{}_f{}.txt".format(_loc, beta)), 'w')

            for tool, new_thresh in final_thresholds.items():

                # delete original threshold
                _new_thresh = new_thresh[1:]
                thresholds_to_return[beta][_loc][tool] = _new_thresh

                t = _new_thresh[i][0]
                # if ci are present, just report thrsh
                if isinstance(t, str):
                    t = t.split()[0]
                info = [v for v in threshold_list if v[0] == tool][0]
                vcf_field = ','.join(tools_config[tool][0])
                outline = [tool, vcf_field, info[1], str(t), info[3]]
                config.write('\t'.join(outline) + '\n')

    return thresholds_to_return


from typing import List
from collections import defaultdict
import sys
import logging
from preprocessing.location import CONSEQUENCES

def update_thresholds(config_dict: defaultdict):
    """
    Updates the list of reference
    thresholds based in the tools
    config file provided and parsed.
    Returns the available methods that exist
    by default in VETA, as well as custom
    methods provided by the user.

    :return List: Complete list of reference
    threshold
    """
    _default_list = [
        ('GERP', '>', 4.4, 'Conservation'),
        ('phyloP', '>', 1.6, 'Conservation'),
        ('SiPhy', '>', 12.17, 'Conservation'),
        ('phastCons', '>', 0.99, 'Conservation'),
        ('CDTS', '<', 10, 'Conservation'),
        ('fitCons', '>', 0.4, 'Whole_genome'),
        ('LINSIGHT', '>', 0.056, 'Whole_genome'), # S-CAP paper

        ('Sift', '<', 0.001, 'Protein'),
        ('Polyphen2HVAR', '>', 0.978, 'Protein'),
        ('Polyphen2HDIV', '>', 0.978, 'Protein'),
        ('MutationAssessor', '>', 1.935, 'Protein'),
        ('MutationTaster2', '>', 0.5, 'Protein'),
        ('FATHMM', '<', -4.14, 'Protein'),
        ('Provean', '<', -2.5, 'Protein'),
        ('LRT', '<', 0.01, 'Protein'),
        ('Mutpred', '>', 0.5, 'Protein'),
        ('VEST4', '>', 0.764, 'Protein'),
        ('CAROL', '>', 0.98, 'Protein'),
        ('Condel', '>', 0.468, 'Protein'),
        ('REVEL', '>', 0.644, 'Protein'),
        ('MetaLR', '>', 0.5, 'Protein'),
        ('MetaSVM', '>', 0.5, 'Protein'),
        ('M-CAP', '>', 0.025, 'Protein'),
        ('MVP', '>', 0.7, 'Protein'),
        ('MTR', '<', 0.5, 'Protein'), # Did not find in the paper, assumed 0.5
        ('MPC', '>', 1.36, 'Protein'),
        ('MISTIC', '>', 0.5, 'Protein'),
        ('CardioBoost', '>', 0.9, 'Protein'),
        ('CardioVAI', '>', 2, 'Protein'),
        ('PrimateAI', '>', 0.790, 'Protein'),
        ('VARITY', '>', 0.75, 'Protein'), # VARITY_R model
        ('ClinPred', '>', 0.5, 'Protein'),
        ('MutScore', '>', 0.5, 'Protein'), # Did not find in the paper, assumed 0.5
        ('MutFormer', '>', 0.5, 'Protein'), # Did not find in the paper, assumed 0.5

        ('cVEP', '>', 0.5, 'Protein'), # 0.5 as artificial threshold
        ('EVE_class20', '>', 0.5, 'Protein'), # 0.5 as artificial threshold
        ('EVE_class50', '>', 0.5, 'Protein'), # 0.5 as artificial threshold
        ('EVE_class90', '>', 0.5, 'Protein'), # 0.5 as artificial threshold
        ('EVE', '>', 0.5, 'Protein'),

        ('CAPICE', '>', 0.02, 'Whole_genome'),
        ('CADD_v1.5', '>', 15, 'Whole_genome'),
        ('CADD-Splice', '>', 15, 'Whole_genome'),

        ('DANN', '>', 0.9, 'Whole_genome'),
        ('GWAVA', '>', 0.5, 'Whole_genome'),
        ('FATHMM-MKL', '>', 0.5, 'Whole_genome'),
        ('Eigen', '>', 4.86, 'Whole_genome'),  # S-CAP paper
        ('ReMM', '>', 0.984, 'Whole_genome'),
        ('FunSeq2', '>', 1.5, 'Whole_genome'),

        # HAL model scores alt5 PSI. kipoi veff scores the DIFF
        # between the ALT and REF allele, thus I set 5 as the
        # minimum threshold to account for PSI changes caused
        # by variant
        ('HAL', '>', 5, 'Splicing'),
        ('MMSplice', '>', 2, 'Splicing'),
        # ('mmsplice-deltaLogitPSI', '>', 2, 'Splicing'),
        # ('mmsplice-pathogenicity', '>', 0.5, 'Splicing'),
        # ('mmsplice-efficiency', '>', 1, 'Splicing'),
        ('kipoiSplice4', '>', 0.5, 'Splicing'),
        ('kipoiSplice4_cons', '>', 0.5, 'Splicing'),

        # S-CAP model has region specific thresholds.
        # To generalize, if score is greater than its
        # specific threshold, I sum 1 to the score
        # (see process_scap_scores method)
        ('ConSpliceML', '>', 0.5, 'Splicing'),
        ('S-CAP', '>', 1, 'Splicing'),
        # TraP, same as S-CAP
        ('TraP', '>', 1, 'Splicing'),
        ('SpliceAI', '>', 0.2, 'Splicing'),
        # SPIDEX: dpsi zscore; dpsi_maxtissue >=5
        # is used in the paper.
        ('SPIDEX', '>', 2, 'Splicing'),
        # Both ada_score and rf_score must be higher
        # than 0.6. Check dbscSNV_merge method on how
        # I dealtwith exceptions
        ('dbscSNV', '>', 0.6, 'Splicing'),
        # MaxEntScan - naive estimation of prediction score.
        # It refers to those variants where the difference in
        # the maximum entropy estimation between the REF and ALT
        # for splice site detection is higher than 3, be it for
        # the gain (positive entropy value) or for the loss
        # (negative entropy value).
        ('MaxEntScan', '>', 3, 'Splicing'),
        ('SQUIRLS', '>', 0.074, 'Splicing'), #CI-SpliceAI paper
        ('IntSplice2', '>', 0.5, 'Splicing'),
        ('CI-SpliceAI', '>', 0.190, 'Splicing'),
        ('Pangolin', '>', 0.2, 'Splicing'),
        ('SPiP', '>', 0.452, 'Splicing'),
        ('MLCsplice', '>', 0.5, 'Splicing'),
        ('AbSplice-DNA', '>', 0.01, 'Splicing'),
        ('LaBranchoR', '>', 0.1, 'Splicing')
    ]

    DEFAULT_TOOLS = [t[0] for t in _default_list]
    _updated_list = []
    for _tool, info in config_dict.items():

        # If user provided custom thresholds
        # for tools available in VETA by
        # default
        if _tool in DEFAULT_TOOLS and len(info) > 1:
            logging.info("Warn: Threshold information about "
                         "{} tool already exists by default "
                         "in VETA. This threshold will be "
                         "updated to the one provided in this run:"
                         "\'{}\'.".format(_tool, info))

            updated_info = tuple([_tool] + info[1:])
            _updated_list.append(updated_info)

        # Use default setting for a
        # known tool
        elif _tool in DEFAULT_TOOLS:
            _default_info = [v for v in _default_list if v[0] == _tool][0]
            _updated_list.append(_default_info)

        # If a custom method
        else:
            if len(info) <= 4:
                custom = [_tool] + info[1:]
            # If custom function, remove it
            elif len(info) == 5:
                custom = [_tool] + info[1:-1]
                
            _updated_list.append(tuple(custom))

    return _updated_list

def _extract_possible_filters(aggregate: bool):
    """
    Extracts possible location/class filters
    to perform an analysis, given the aggregate
    option provided
    
    :return list: List with lambda function to filter df by each loc
    """
    idx = 0 if aggregate else 1
    all_locs = list(set([x[idx] for x in CONSEQUENCES.values()]))
    all_locs.append('deep_intronic')
    all_locs.insert(0, 'all')
    return all_locs

def subset_variants_by_type(types_to_analyse: List = None):
    """
    Subset filters for variant types
    :param List types_to_analyse: List of types
    to filter. Default: `None`
    :return List: Variant types that will be used
    """
    if types_to_analyse is None:
        return filters_var_type
    else:
        return [var_types for var_types in filters_var_type if var_types[0] in types_to_analyse]


def subset_toolset_by_scope(threshold_list: List,
                            scopes: List = None):

    """Returns subset of tools belonging to the
    scope defined, or if `is_intronic` is true,
    removes tools belonging to `Protein scope,
    regardless of the `scores_to_analys` value

    :param List threshold_list: List of reference
        thresholds defined taking into account the
        default thresholds in VETA as well as the custom
        ones provided by the user

    :param List scopes: Tool scopes to analyse.
        Default: `None`. Use all scopes
    """

    _to_analyse = ['Conservation', 'Protein', 'Whole_genome', 'Splicing']

    if scopes is not None:
        _to_analyse = list(set(_to_analyse).intersection(scopes))

    out = []
    for tool in threshold_list:
        try:
            if tool[3] in _to_analyse:
                out.append(tool)
        except IndexError:
            raise IndexError('Scope is missing in the config file for {} tool'.format(tool))
    return out


filters_var_type = [
    ('all_types', lambda x: x),
    ('snps', lambda x: x[x['type'].str.match('snp')]),
    ('indels', lambda x: x.query('type == "indel" & subtype == "ins" | type == "indel" & subtype == "del"')),
    ('insertions', lambda x: x.query('type == "indel" & subtype == "ins"')),
    ('deletions', lambda x: x.query('type == "indel" & subtype == "del"')),
    ('mnps', lambda x: x.query('type == "indel" & subtype == "mnp"'))
]

filter_intronic_bins = [
    ('all_intronic', lambda x: x[~x['intron_bin'].isnull()]),
    ('all_except_1-2', lambda x: x[~x['intron_bin'].str.match('1-2')]),
    ('all_except_1-10', lambda x: x[(~x['intron_bin'].str.match('1-2')) &
                                    (~x['intron_bin'].str.match('3-10')) & 
                                    (~x['intron_bin'].str.match('1-10'))]),
    ('1-2', lambda x: x[x['intron_bin'].str.match('1-2')]),
    ('3-10', lambda x: x[x['intron_bin'].str.match('3-10')]),
    ('1-10', lambda x: x[x['intron_bin'].str.match('1-10')]),
    ('11-40', lambda x: x[x['intron_bin'].str.match('11-40')]),
    ('41-200', lambda x: x[x['intron_bin'].str.match('41-200')]),
    ('201-500', lambda x: x[x['intron_bin'].str.match('201-500')]),
    ('501-1000', lambda x: x[x['intron_bin'].str.match('501-1000')]),
    ('1000+', lambda x: x[x['intron_bin'].str.match('1000\+')])
]